import numpy as np
from lmfit import minimize, Parameters, report_fit
from functools import partial
import multiprocessing as mp
import copy
import itertools

#constants
gas_const_kcal = .0019872036

class datafit(object):
    """extend with fxn_index  methods. index must be an integer.

    Docstring text (first line) is parsed for building function table so docstring format must be conserved

        Ex1: Generic Function: Constant (Additive)
            \n
        Ex2: Generic Function: Constant (Additive)\n

        where "Generic Function" is the category of the function and "Constant (additive)" is the function name"

    Remaining doctstring format should be conserved as it is the description used by the 'fun info' command.
        """

    def __init__(self, data_instance, info_commands=('info', '?')):
        self._info_cmd = info_commands
        self.paramid = []
        self.paramval = []
        self.parambounds = []
        self.paramdefaults = []
        self.funcindex = []
        self.functions = []
        self.pyscript = False
        self.tzerooffset = False
        self.pyscriptonly = False
        self.inst = data_instance

    def __call__(self, args):
        if len(args) == 0:
            return 'error'
        elif len(args) == 1 and args[0] == "info":
            return self.info(None)
        elif len(args) > 1 and args[0] == "info":
            return self.info(args[1])

    def update(self, funcnum):
        for num in funcnum:
            fn = getattr(self, 'fxn_' + str(num))
            fn()
        return True

    def clear(self):
        self.paramid = []
        self.paramval = []
        self.parambounds = []
        self.paramdefaults = []
        self.funcindex = []
        self.functions = []
        self.pyscript = []
        return True

    def fxn_1(self, *args):
        """Generic Function: Constant (Additive)
        \nDescription: Function in the form of: Y = C + 0X
        \nParameters:
        \tC\t(constant value)
        """
        self.paramid.extend(["Add_Constant"])
        self.parambounds.extend([[-np.inf, np.inf]])
        self.paramdefaults.extend([0])
        self.functions.extend(["Y=(X*0)+P[0]"])
        return

    def fxn_2(self, *args):
        """Generic Function: Exponential
        \nDescription: Function in the form of: Y = A*e^(-X/t)
        \nParameters:
        \tA\t(exponential amplitude)
        \tt\t(exponential time constant)
        """
        self.paramid.extend(["Amplitude", "Time_Constant"])
        self.parambounds.extend([[-np.inf, np.inf], [0, np.inf]])
        self.paramdefaults.extend([1, 1])
        self.functions.extend(["Y=P[0]*np.exp(-1*X/P[1])"])
        return

    def fxn_3(self, *args):
        """Generic Function: Gaussian 1-D
        \nDescription: Function in the form of: Y = A/(sqrt(2*pi)*WHM))*exp(-(X-CENTER)^2/(2*WHM^2)
        \nParameters:
        \tA   \t(Amplitude)
        \tWHM \t(Full Width at Half Max)
        \tXcen\t(Centered at X-val)
        """
        self.paramid.extend(["Amplitude", "WHM", "Xcen"])
        self.parambounds.extend([[-np.inf, np.inf], [0, np.inf], [-np.inf, np.inf]])
        self.paramdefaults.extend([10, 5, 5])
        self.functions.extend(["Y=(P[0]/(np.sqrt(2*np.pi)*P[1]))*np.exp(-(X-P[2])**2/(2*P[1]**2))"])
        return

    def fxn_27(self, *args):
        """Generic Function: Linear
        \nDescription: Function in the form of: Y = M*X+C
        \nParameters:
        \tM\t(Mul. constant value [slope])
        \tC\t(Add. constant value [y-intercept])
        """
        self.paramid.extend(["Slope", "Y_intercept"])
        self.parambounds.extend([[-np.inf, np.inf], [-np.inf, np.inf]])
        self.paramdefaults.extend([1, 0])
        self.functions.extend(["Y=(P[0]*X)+P[1]"])
        return

    def fxn_30(self, *args):
        """Generic Function: Constant (Multiplicative)
        \nDescription: Function in the form of: Y = M*X
        \nParameters:
        \tM\t(constant value)
        """
        self.paramid.extend(["Mul_Constant"])
        self.parambounds.extend([[-np.inf, np.inf]])
        self.paramdefaults.extend([0])
        self.functions.extend(["Y=X*P[0]"])
        return

    def fxn_14(self, *args):
        """Protein Folding Equilibrium: 2-state Equilibrium, Chemical Denaturant
        \nDescription: Model for equilibrium titration data describing 2 thermodynamic states (U->N)
        \nPubMed ID: 15689503
        \nParameters:
        \tdG \t(Delta G in kcal/Mol)
        \tm  \t(m-value, denaturant dependence of transition in kcal/mol/M)
        \tCn \t(Y-intercept of Native Baseline)
        \tCu \t(Y-intercept of Unfolded Baseline)
        \tMn \t(Slope of Native Baseline)
        \tMu \t(Slope of Unfolded Baseline)
        \tT  \t(Temperature in Kelvin)

        Note: This function requires the X-values to be denaturant concentrations and the Y-values are the experimental signal at that denaturant concentration
        """
        self.paramid.extend(["dG", "m", "Cn", "Cu", "Mn", "Mu", "T"])
        self.parambounds.extend([[-np.inf, np.inf], [-np.inf, np.inf], [-np.inf, np.inf], [-np.inf, np.inf],
                                 [-np.inf, np.inf], [-np.inf, np.inf], [-np.inf, np.inf]])
        self.paramdefaults.extend([5, 1.8, 0, 25000, 1000, 2, 298.15])
        self.functions.extend(["Y=(P[2]+(P[4]*X))+((P[3]+(P[5]*X))*np.exp((P[0]-(P[1]*X))/(gas_const_kcal*P[6])))/(1+np.exp((P[0]-(P[1]*X))/(gas_const_kcal*P[6])))"])
        self.pyscriptonly = False
        return

    def fxn_39(self, *args):
        """Molecular Kinetics Function: CFCA (SPR, Biacore 8K)
        \nDescription: Calibration-Free Concentration Analysis for SPR data. Requires 2 datasets of concentration series
         data collected at different flow rates. One data set should have kinetics that are at least
         partially limited by mass transport. Calculation assumes Biacore 8K flow cell.
        \nPubMed ID: 12081475
        \nParameters:
        \tFd  \t(Fold Dilution from max concentration in series)
        \tF\t \t(Flow Rate in uL/min)
        \tMW  \t(Molecular Weight of Analyte in Da)
        \tT   \t(Temperature of experiment in Celcius)
        \tRmax\t(Maximum SPR Response)
        \tka  \t(Association Rate Constant in 1/Ms)
        \tkd  \t(Dissociation Rate Constant in 1/s)
        \tCa  \t(Concentration of Analyte in M)
        \tt0  \t(Time Zero of Dissociation Data in s)
        \tkc  \t(Mass Transport Coefficient in cm/s.  If user enters 0, estimation with Einstein-Sutherland equation overrides.)
        \tG   \t(Concentration G-Factor in Rcm^2/g. If user enters 0, Defaults to 100000 Rmm^2/ng)
        """
        self.paramid.extend(["Fd", "F", "MW", "T", "Rmax", "ka", "kd", "Ca", "t0", "kc", "G"])
        self.parambounds.extend([[0, np.inf], [0, np.inf], [0, np.inf], [0, np.inf], [0, np.inf], [0, np.inf],
                                 [0, np.inf], [0, np.inf], [-np.inf, np.inf], [0, np.inf], [0, np.inf]])
        self.paramdefaults.extend([1, 30, 14300, 25, 25, 2.3E6, 9.85E-11, 5E-9, 300, 0, 1.0E10])
        self.pyscriptonly = True
        self.pyscript = compile("""
# L1 is the distance from the inlet to the start of the detection area
# L2 is the distance from the inlet to the end of the detection spot
# SpotL and SpotW are the detection spot dimensions
#dimensions are in cm
SpotL = .16
SpotW = 0.02
L1 = .03
L2 = .03 + SpotL
# Flow Cell Dimensions in mm from Christensen paper
CW = 0.05
CH = 0.005
CL = 0.24
# CKC describes the concentration gradient
CKC = 1.47 * ((1 - np.power((L1 / L2), 0.6666666667)) / (1 - (L1 / L2)))
# D is the Diffusion coefficient of analyte estimated using the Einstein-Sutherland equation (christensen paper)
D = (1.381E-23 * (298.15 + P[3])) / (6 * np.pi * np.power((3 * np.pi * P[2] * (7.3E-4 / (4 * np.pi * 6.022E23))), 0.3333333333) * 0.001 * 1.2)
# if user enters 0 for kc, estimate it here:
P[7] = (P[7]/P[0]) #corrects concentration for fold dilution
P[2] = P[2] #MW is in Da
P[1] = (P[1]*60) #converts flowrate from uL/min to uL/sec
P[3] = P[3] + 273.15 #converts C to Kelvin
if P[9] == 0:
    P[9] = CKC * np.power((((D * D) * P[1]) / (np.power(CH,2) * CW * L2)), 0.3333333333)
K1 = (P[7] * P[5]) / ((P[7] * P[5]) + P[6])
K2 = (P[7] * np.power(P[5],2) * P[4]) / ((((P[7] * P[5]) + P[6]) * P[9] * P[2] * P[10]) + (P[5] * P[6] * P[4]))
K3 = (np.power(((P[7] * P[5]) + P[6]), 2) * P[9] * P[2] * P[10]) / (
P[6] * ((((P[7] * P[5]) + P[6]) * (P[9] * P[2] * P[10])) + (P[5] * P[6] * P[4])))
K4 = ((P[2] * P[10] * P[9]) + (P[5] * P[4])) / (P[5] * P[4])
K5 = (P[2] * P[10] * P[9]) / ((P[2] * P[10] * P[9]) + (P[5] * P[4]))
K6 = ((P[2] * P[10] * P[9]) + (P[5] * P[4])) / P[5]
Rz = None
for j in range(len(X)):
    if X[j] <= P[8]:
        # calc association
        Wval = numericalmethods.lambertw(K2 * np.exp(K2 - (K3 * P[6] * X[j])))
        R[j] = P[4] * K1 * (1 - (Wval / K2))
        Rz = float(R[j])
    else:
        # calc dissociation
        Wval = numericalmethods.lambertw(((-1 * Rz) / K6) * np.exp(-1 * ((Rz / K6) + (K5 * P[6] * (X[j] - P[8])))))
        R[j] = (-1 * K6) * Wval
""", '<string>', 'exec')
        self.functions.extend(["Y=X+P[0]"])
        self.pyscriptonly = True
        return

    def fxn_40(self, *args):
            """Molecular Kinetics Function: 1-to-1 Stoiciometery binding model (on & off, same buffer)
            \nDescription: 1:1 Stoiciometery binding model for molecular interactions. Full trace.
            \nPubMed ID: 28430560
            \nParameters:
            \tRmax\t(Response maximum value)
            \tkd  \t(Dissociation rate)
            \tka  \t(Association rate)
            \tCp  \t(Concentration of analyte in solution in Molar)
            \tm   \t(Linear approximation of slow phase)
            \tc   \t(Dissociation asymtote)
            \tX0  \t(Dissociation phase time offset)
            \tkds \t(Dissociation phase scalar)
            """
            self.paramid.extend(["Rmax", "kd", "ka", "Cp", "m", "c", "X0", "kds"])
            self.parambounds.extend([[0, np.inf], [0, np.inf], [0, np.inf], [0, np.inf], [-np.inf, np.inf],
                                     [-np.inf, np.inf], [-np.inf, np.inf], [-6, 6]])
            self.paramdefaults.extend([25, 0.001, 60000, 100E-9, 0, 0, 180, 1])
            self.pyscriptonly = True
            self.pyscript = compile("""
rm, kd, ka, cp, m, c, xo, kds = P
if cp==0:
    cp=1E-15
for j in range(len(X)):
    if X[j] <= xo:
        # calc association
        R[j] = (rm / (1 + (kd / (ka * cp)))) * (1-(np.exp((-1 * X[j] * ((ka * cp) + kd)))))
    else:
        # calc dissociation
        R[j] = kds*(((rm/(1+(kd/(ka*cp))))*(1-np.exp(-1*xo*(ka*cp+kd)))-c))*(np.exp(-1*kd*(X[j]-xo))) + m*(X[j]-xo) + c
""", '<string>', 'exec')

            self.functions.extend(["Y=(P[0]/(1+(P[1]/(P[2]*P[3]))))*(1-(np.exp((-1*X*((P[2]*P[3])+P[1])))))+(P[7]*(((P[0]/(1+(P[1]/(P[2]*P[3]))))*(1-np.exp(-1*P[6]*(P[2]*P[3]+P[1])))-P[5]))*(np.exp(-1*P[1]*(X-P[6])))+P[4]*(X-P[6])+P[5])"])
            return

    def fxn_41(self, *args):
        """Molecular Equilibrium Binding Function: 1-to-1 Stoiciometery binding model using Hill Equation
        \nDescription: 1:1 Stoiciometery binding model for molecular interactions. Hill equation estimate of EC50.
        \nPubMed ID: 19049668
        \nParameters:
        \tEC50    \t(Effective concentrarion for 50% binding)
        \tHillCoef\t(Hill coefficient)
        \tminY    \t(Minimum Y-value)
        \tmaxY    \t(Maximum Y-value)
        """
        self.paramid.extend(["EC50", "HillCoef", "minY", "maxY"])
        self.parambounds.extend([[0, np.inf], [-1000, 1000], [-np.inf, np.inf], [-np.inf, np.inf]])
        self.paramdefaults.extend([5, 1, 0, 100])
        self.functions.extend(["Y=P[2]+((P[3]-P[2])/(1+((P[0]/X)**P[1])))"])
        # Y=minY+((maxY-minY)/(1+((EC50/X[0])^HillCoef)))
        return

    def fxn_42(self, *args):
        """Molecular Equilibrium Binding Function: 1-to-1 Stoiciometery binding model using Quadratic Equation
        \nDescription: 1:1 Stoiciometery binding model for molecular interactions. Quadratic equation estimate of Keq.
        \nPubMed ID: 21115850
        \nParameters:
        \tKeq    \t(Equilibrium dissociation constant)
        \tC      \t(Concentration of constant component)
        \tAmp    \t(Signal Amplitude [max_Y - min_Y])
        \tS0     \t(background signal of unbound [min_Y])
        """
        self.paramid.extend(["Keq", "C", "Amp", "S0"])
        self.parambounds.extend([[0, np.inf], [0, np.inf], [-np.inf, np.inf], [-np.inf, np.inf]])
        self.paramdefaults.extend([0.5, 100, 1000, 0])
        self.functions.extend(["Y=(P[3]+P[2])*((P[1]+X+P[0])-np.power(np.power(P[1]+X+P[0], 2)-(4*P[1]*X), 0.5)) /(2*P[1])"])
        # Y=S0+Amp*((c+X+Keq)-((c+X+Keq)^2 -(4*c*X))^0.5 )/2*c
        return

    def fxn_43(self, *args):
        """Molecular Equilibrium Binding Function: 1-to-1 Stoiciometery Single Site Specific Binding Potential
        \nDescription: 1:1 Stoiciometery binding potential model for molecular interactions. Binding Potential estimate of Keq.
        \nPubMed ID: 6609679
        \nParameters:
        \tKeq    \t(Equilibrium dissociation constant)
        \tBmax   \t(Signal of maximum binding [max_Y])
        \tS0     \t(background signal of unbound [min_Y])
        \nNote (binding potential (BP)):
        \tBP     \t(Bmax/Keq = receptor density * affinity)
        """
        self.paramid.extend(["Keq", "Bmax", "S0"])
        self.parambounds.extend([[0, np.inf], [0, np.inf], [-np.inf, np.inf]])
        self.paramdefaults.extend([0.5, 1000, 0])
        self.functions.extend(["Y=((P[1]*X)/(P[0]+X))+P[2]"])
        # Y=(Bmax*X/(Keq + X))+S0; BP=Bmax/Keq
        return

    def info(self, fxn=None):
        def std_info():
            hc = '|'.join(self._info_cmd)
            res = '\n\tType [%s] fxn_index to get more help about particular command\n' % hc
            fl = self.getfxnlist()
            res += '\n\tAvailable Functions: \n\t%s' % ('  '.join(sorted(fl))) + "\n"
            return res

        if not fxn:
            return std_info()
        else:
            try:
                fn = getattr(self, 'fxn_' + str(fxn))
                doc = fn.__doc__
                return doc or 'No documentation available for %s' % fxn
            except AttributeError:
                return std_info()

    def getfxnlist(self):
        return [name[4:] for name in dir(self) if name.startswith('fxn_') and len(name) > 4]

    def showfxntable(self):
        topline = 100*'_'
        spacer = 86*'.'
        toreturn = "\n\t     Function Name" + 60*' ' + "Function Number\n" + topline + '\n'
        cat = []
        fxnindex = [name[4:] for name in dir(self) if name.startswith('fxn_') and len(name) > 4]
        name = []
        for fxn in fxnindex:
            fn = getattr(self, 'fxn_' + str(fxn))
            doc = fn.__doc__
            doc = doc.split('\n')
            doc = doc[0].split(':')
            cat.append(doc[0].strip())
            name.append(doc[1].strip())
        titles = list(set(cat))
        for t in titles:
            toreturn += (t + ':\n')
            for i in range(len(cat)):
                if cat[i] == t:
                    toreturn += ('\t' + name[i] +
                                 spacer[:len(spacer)-(len(name[i])+len(fxnindex[i]))] + fxnindex[i] + '\n')
        return toreturn

    def applyfxns(self):
        firstindex = 1
        lastindex = 1
        fxnindicies = self.funcindex
        fxns = self.functions
        for f in self.funcindex:
            fn = getattr(self, 'fxn_' + str(f))
            fn()
        if self.inst.data.plot_limits.is_active:
            firstindex = min(self.inst.data.plot_limits.buffer_range.get())
            lastindex = max(self.inst.data.plot_limits.buffer_range.get())
        else:
            firstindex = 1
            lastindex = self.inst.data.matrix.length() + 1
        maxindex = 0
        for k in range(len(fxns)):
            fxnsplit = fxns[k].split("P[")
            for f in range(0, len(fxnsplit) - 1, 1):
                fxnsplit[f] += "P["
                tempstr = fxnsplit[f + 1].split(']')
                fxnsplit[f + 1] = str(int(tempstr[0]) + maxindex) + ']' + tempstr[1]
            maxindex = int(tempstr[0]) + maxindex + 1
            fxns[k] = ''.join(fxnsplit)
        tempfxn = ''
        for j in range(len(fxns)):
            yval, xval = fxns[j].split('=')
            tempfxn += xval + '+'
        try:
            for i in range(firstindex, lastindex):
                self.inst.data.matrix.buffer(i).fit.function_index.set(fxnindicies)
                self.inst.data.matrix.buffer(i).fit.function.set(tempfxn[:-1])
        except:
            return False
        return True

    def dofit(self, *args):
        fitparams = datafit(self.inst)
        args = [int(val) if val.isdigit() else val.lower() for val in args]
        method = "Leastsq"
        debug = True if "-debug" in args else False
        group = abs(args[args.index('-group') + 1]) if '-group' in args and args.index('-group') < len(args) and is_integer(args[args.index('-group') + 1]) else 1 if '-ind' in args else self.inst.data.matrix.length()
        cpu = abs(args[args.index('-cpu') + 1]) if '-cpu' in args and args.index('-cpu') < len(args) and is_integer(args[args.index('-cpu') + 1]) else 1
        ind_fit = True if '-ind' in args else False
        group = 1 if ind_fit else group
        silent = True if "-silent" in args else False
        iter_cb = debug_fitting if debug else None
        max_iter = int(args[0]) if isinstance(args[0], int) else 2000
        bmax = self.inst.data.matrix.length() #if not self.inst.data.plot_limits.is_active else max(self.inst.data.plot_limits.buffer_range.get())
        bmin = 1 if not self.inst.data.plot_limits.is_active else min(self.inst.data.plot_limits.buffer_range.get())

        #construct list of lists for X, Y, Params
        X_vec = []
        Y_vec = []
        Z_vec = []
        P_vec = []
        IR_X_vec = []
        IR_Y_vec = []
        IR_Z_vec = []
        FXN_vec = []
        WEIGHTS_vec = []
        FXN_NUM_vec = []
        PARAM_ID_vec = []
        Y_matrix = []
        for i in range(bmin, bmax+1, group):
            parameters = Parameters()
            x_group = []
            y_group = []
            z_group = []
            ir_x_group = []
            ir_y_group = []
            ir_z_group = []
            fxn_group = []
            weights_group = []
            fxn_num_group = []
            for j in range(group):
                k=i+j
                fitparams.clear()
                fitparams.update(self.inst.data.matrix.buffer(k).fit.function_index.get())
                p_init= self.inst.data.matrix.buffer(k).fit.parameter.get()
                if len(p_init) == 0:
                    return "Invalid Parameters!  Try Function: ap ."

                for m in range(len(p_init)):
                    try:
                        parameters.add(name=fitparams.paramid[m] + "_{}_{}".format(m+1, k),
                                       value=float(self.inst.data.matrix.buffer(k).fit.parameter.get()[m]),
                                       min=float(min(fitparams.parambounds[m])), max=float(max(fitparams.parambounds[m])),
                                       expr=self.inst.data.matrix.buffer(k).fit.link.get()[m],
                                       vary=self.inst.data.matrix.buffer(k).fit.free.get()[m])
                    except (NameError, ValueError) as e:
                        if isinstance(e, NameError):
                            return "Parameter Linking Scheme is Invalid!"
                        elif isinstance(e, ValueError):
                            return "Parameters Return Invalid Results!!"

                x_group.append(self.inst.data.matrix.buffer(k).data.x.get())
                y_group.append(self.inst.data.matrix.buffer(k).data.y.get())
                z_group.append(self.inst.data.matrix.buffer(k).data.z.get())
                ir_x_group.append(self.inst.data.matrix.buffer(k).instrument_response.x.get())
                ir_y_group.append(self.inst.data.matrix.buffer(k).instrument_response.y.get())
                ir_z_group.append(self.inst.data.matrix.buffer(k).instrument_response.z.get())
                weights_group.append(self.inst.data.matrix.buffer(k).data.ye.get())
                fxn_group.append(self.inst.data.matrix.buffer(k).fit.function.get())
                temp_df = datafit(self.inst)
                temp_df.update(self.inst.data.matrix.buffer(k).fit.function_index.get())
                fxn_num_group.append(self.inst.data.matrix.buffer(k).fit.function_index.get())
            X_vec.append(x_group)
            Y_vec.append(y_group)
            Z_vec.append(z_group)
            P_vec.append(parameters)
            IR_X_vec.append(ir_x_group)
            IR_Y_vec.append(ir_y_group)
            IR_Z_vec.append(ir_z_group)
            WEIGHTS_vec.append(weights_group)
            FXN_vec.append(fxn_group)
            FXN_NUM_vec.append(fxn_num_group)

        try:
            for vec in Y_vec:
                Y_matrix.append(np.array(vec))
                assert Y_matrix[-1].shape == (len(Y_matrix[-1]), max([len(y) for y in vec]))
        except AssertionError:
            return "All Buffers Must Be the Same Number of Points!  Try Commands: pl or res or tri"

        param_dict = {'x_vec': X_vec, 'y_vec': Y_vec, 'z_vec': Z_vec, 'p_vec':P_vec, 'y_matrix': Y_matrix,
                      'ir_x_vec': IR_X_vec, 'ir_y_vec': IR_Y_vec, 'ir_z_vec': IR_Z_vec, 'silent': silent,
                      'weights_vec': WEIGHTS_vec, 'fxn_vec': FXN_vec, 'fxn_num_vec': FXN_NUM_vec,
                      'param_id_vec': PARAM_ID_vec,'method': method, 'debug': debug, 'group': group, 'cpu': cpu,
                      'ind_fit': ind_fit, 'iter_cb': iter_cb, 'max_iter': max_iter, 'index': i}

        result = multi_fit(param_dict)
        result = self.split_result_by_group(result, group)

        print('Saving parameters to matrix...')
        self.saveparams(result, group, silent)

        print('Calculating fit stats and generating model traces...')
        for i in range(bmin, bmax+1):
            try:
                self.calcfitstat(i)
                self.generatemodel(i, numpts=300)
            except Exception as e:
                print(str(e))

            if debug:
                report_fit(result[i-bmin].params)

            if not silent:
                print(f'---Buffer {i} fit statistics---')
                # print number of function efvals
                print('\n#Function efvals:\t', result[i-bmin].nfev)
                #print number of data points
                print('#Data pts:\t', result[i-bmin].ndata)
                #print number of variables
                print('#Variables:\t', result[i-bmin].nvarys)
                # chi-sqr
                print('\nResult Chi Sq:\t', result[i-bmin].chisqr)
                # reduce chi-sqr
                print('Result Reduced Chi Sq:\t', result[i-bmin].redchi)
                # Akaike info crit
                print('Result Akaike:\t', result[i-bmin].aic)
                # Bayesian info crit
                print('Result Bayesian:\t', result[i-bmin].bic)
                # message
                print('Fit Details:\t', result[i-bmin].message)
                print('-----------------------------')
        return "\nData Fitting Complete!"

    def split_result_by_group(self, result, group):
        '''Make result vector equivilent size to buffer matrix by splitting results by group'''
        if group == 1:
            return result
        result_to_return = []
        for r in result:
            for i in range(group):
                if r == None:
                    result_to_return.append(None)
                    continue
                result_to_return.append(copy.copy(r))
        return result_to_return

    def generatemodel(self, i, numpts=300):
        X = self.inst.data.matrix.buffer(i).data.x.get()
        Y = self.inst.data.matrix.buffer(i).data.y.get()
        Z = self.inst.data.matrix.buffer(i).data.z.get()
        IRX = self.inst.data.matrix.buffer(i).instrument_response.x.get()
        IRY = self.inst.data.matrix.buffer(i).instrument_response.y.get()
        IRZ = self.inst.data.matrix.buffer(i).instrument_response.z.get()

        min_x = np.min(X)
        max_x = np.max(X)
        stepsize = (max_x - min_x) / (numpts)
        modelX = [((stepsize * k) + min_x) for k in range(0, numpts, 1)]

        # add some points in-case data is logarithmically sampled 11/08/2019
        inc_x = list(zip(['inc'] * len(modelX), modelX))
        data_x = list(zip(['data'] * len(X), X))
        inc_x.extend(data_x)
        all_x = sorted(inc_x, key=lambda tup: tup[1])
        for j in reversed(range(1, len(all_x) - 1, 1)):
            if all_x[j][0] == 'data' and all_x[j - 1][0] == 'inc' and all_x[j + 1][0] == 'inc':
                del all_x[j]
        _, X = map(list, zip(*all_x))
        # end add points

        fxn = self.inst.data.matrix.buffer(i).fit.function.get()
        P = self.inst.data.matrix.buffer(i).fit.parameter.get()
        X = np.array(X)
        R = [0] * len(X)

        if fxn is not False and fxn[:2].upper() == "Y=":
            fxn = fxn[2:]
        # execute custom script
        if self.pyscript:
            exec(self.pyscript)
        else:
            R = eval(str(fxn))

        self.inst.data.matrix.buffer(i).model.x.set(X)
        self.inst.data.matrix.buffer(i).model.y.set(R)
        return True

    def calcfitstat(self, i):
        rawy = self.inst.data.matrix.buffer(i).data.y.get()
        resid_y = self.inst.data.matrix.buffer(i).residuals.y.get()
        rsq = np.sum(np.power(resid_y, 2))
        avgerror = np.sum(np.power(rawy - np.average(rawy), 2))
        avgerror = 1E-6 if avgerror == 0 else avgerror
        self.inst.data.matrix.buffer(i).fit.rsq.set(1 - (rsq / avgerror))
        SD = np.average(np.abs(resid_y))
        if SD == 0:
            SD = 0.001
        self.inst.data.matrix.buffer(i).fit.chisq.set(np.sum(((resid_y) ** 2) / SD))
        return

    def saveparams(self, result, group, silent):
        bmin = 1
        if not self.inst.data.plot_limits.is_active:
            bmax = self.inst.data.matrix.length()
        else:
            bmin = self.inst.data.plot_limits.buffer_range.min()
            bmax = self.inst.data.plot_limits.buffer_range.max()

        if not silent:
            print('--------Fit Parameters---------')

        grp_cnt = -1
        for i in range(bmin, bmax+1, 1):
            grp_cnt = grp_cnt + 1 if grp_cnt < group-1 else 0
            i_idx = i-bmin
            self.clear()
            self.update(self.inst.data.matrix.buffer(i).fit.function_index.get())
            self.inst.data.matrix.buffer(i).fit.parameter_error.set([0] * len(self.inst.data.matrix.buffer(i).fit.parameter.get()))
            parameters = self.inst.data.matrix.buffer(i).fit.parameter.get()

            # if fit failed fill in data
            if result[i_idx].aborted:
                x = self.inst.data.matrix.buffer(i).data.x.get()
                y = self.inst.data.matrix.buffer(i).data.y.get()
                z = self.inst.data.matrix.buffer(i).data.z.get()
                self.inst.data.matrix.buffer(i).fit.fit_failed = True
                self.inst.data.matrix.buffer(i).fit.fit_failed_reason.set(str(e))
                self.inst.data.matrix.buffer(i).fit.parameter.set([-1] * len(parameters))
                self.inst.data.matrix.buffer(i).fit.parameter_error.set([-1] * len(parameters))
                self.inst.data.matrix.buffer(i).model.x.set([x[0], x[-1]] if len(x) > 1 else [])
                self.inst.data.matrix.buffer(i).model.y.set([y[0], y[-1]] if len(y) > 1 else [])
                self.inst.data.matrix.buffer(i).model.z.set([z[0], z[-1]] if len(z) > 1 else [])
                self.inst.data.matrix.buffer(i).residuals.y.set([y[0], y[-1]] if len(y) > 1 else [])
                self.inst.data.matrix.buffer(i).residuals.x.set(x)
                if not silent:
                    print(f'Buffer {i}: Fit Failed!\n-----------------------------')
                continue

            # Else, add fit  values to matrix
            resid= np.array_split(result[i_idx].residual, group)
            # weights used are typically Y error vector
            weights = self.inst.data.matrix.buffer(i).data.ye.get()
            # Residuals are squared in fit minimization and multiplied by the weight vector.  Here we reverse those calculations
            try:
                unweighted_resid = np.power(resid[grp_cnt], 0.5) if len(weights) <= 1 else np.power(resid[grp_cnt], 0.5) / weights
            except:
                weights = np.nan_to_num(weights, nan=1.0)
                unweighted_resid = np.power(resid[grp_cnt], 0.5) if len(weights) <= 1 else np.power(resid[grp_cnt], 0.5) / weights
            self.inst.data.matrix.buffer(i).residuals.y.set(unweighted_resid)
            self.inst.data.matrix.buffer(i).residuals.x.set(self.inst.data.matrix.buffer(i).data.x.get())
            for j_idx in range(len(self.paramid)):
                j = j_idx+1
                param = result[i_idx].params[self.paramid[j_idx] + f"_{j}_{i}"].value
                error = result[i_idx].params[self.paramid[j_idx] + f"_{j}_{i}"].stderr
                param_list = self.inst.data.matrix.buffer(i).fit.parameter.get()
                param_list[j_idx] = param
                self.inst.data.matrix.buffer(i).fit.parameter.set(param_list)
                error_list = self.inst.data.matrix.buffer(i).fit.parameter_error.get()
                error_list[j_idx] = error
                self.inst.data.matrix.buffer(i).fit.parameter_error.set(error_list)
                if not silent:
                    print(f'Buffer {i}: Parameter: {self.paramid[j_idx]} = {param} +/- {error}')

            if not silent:
                print('-----------------------------')
        return True


def debug_fitting(self, params, nfev, resid, *args, **kwargs):
    """Function to be called after each iteration of the minimization method
    used by lmfit. Should reveal information about how parameter values are
    changing after every iteration in the fitting routine. See
    lmfit.Minimizer.__residual for more information."""
    print("Iteration {0}".format(nfev) + "\tRsq: " + str(np.sum(np.power(resid, 2))))


def is_integer(input):
    try:
        num = int(input)
    except ValueError:
        return False
    return True

def eval_objective(params, y_matrix, idx, param_dict):  # calculate residuals to determine if the parameters are improving the fit
    '''param_dict = param_dict = {'x_vec': X_vec, 'y_vec': Y_vec, 'z_vec': Z_vec, 'p_vec':P_vec, 'y_matrix': Y_matrix,
                          'ir_x_vec': IR_X_vec, 'ir_y_vec': IR_Y_vec, 'ir_z_vec': IR_Z_vec,
                          'weights_vec': WEIGHTS_vec, 'fxn_vec': FXN_vec, 'fxn_num_vec': FXN_NUM_vec,
                          'param_id_vec': PARAM_ID_vec,'method': method, 'debug': debug, 'group': group, 'cpu': cpu,
                          'ind_fit': ind_fit, 'iter_cb': iter_cb, 'max_iter': max_iter}'''
    resid = 0.0 * y_matrix[:]
    group = param_dict['group']
    x_vec = param_dict['x_vec'][idx]
    y_vec = param_dict['y_vec'][idx]
    z_vec = param_dict['z_vec'][idx]
    p_vec = param_dict['p_vec'][idx]
    ir_x_vec = param_dict['ir_x_vec'][idx]
    ir_y_vec = param_dict['ir_y_vec'][idx]
    ir_z_vec = param_dict['ir_z_vec'][idx]
    fxn_vec = param_dict['fxn_vec'][idx]
    weights_vec = param_dict['weights_vec'][idx]
    fxn_num_vec = param_dict['fxn_num_vec'][idx]
    pyscript_vec = []
    param_id_vec = []
    for fxn_num in fxn_num_vec:
        temp_df = datafit(fxn_num)
        temp_df.update(fxn_num)
        pyscript_vec.append(temp_df.pyscript)
        param_id_vec.append(temp_df.paramid)
    for i in range(len(x_vec)):
        P = []
        X = x_vec[i]
        Y = y_vec[i]
        Z = z_vec[i]
        IRX = ir_x_vec[i]
        IRY = ir_y_vec[i]
        IRZ = ir_z_vec[i]
        weights = weights_vec[i]

        # Fit the data
        fxn = fxn_vec[i]
        R = [0] * len(X)
        for j in range(len(param_id_vec[i])):
            pname = param_id_vec[i][j].replace('-', '') + "_{}_{}".format(j + 1, i+(idx*group)+1)
            P.append(params[pname].value)
        if fxn is not False and fxn[:2].upper() == "Y=":
            fxn = fxn[2:]
        # execute custom script
        if pyscript_vec[i]:
            exec(pyscript_vec[i])
        else:
            R = eval(str(fxn))

        p_vec = P
        R = np.array(R)

        resid_line = (Y - R) if len(weights) <= 1 else (Y - R) * weights
        resid[i, :] = np.power(resid_line, 2)
    # now flatten this to a 1D array, as minimize() needs
    return resid.flatten()

def optimizer(param_dict, idx_list):
    result = []
    for idx in idx_list:
        print(f'Evaluating #{idx+1} from total pool of: {len(param_dict["y_vec"])}')
        parameters = param_dict['p_vec'][idx]
        iter_cb = param_dict['iter_cb']
        max_iter = iter_cb = param_dict['max_iter']
        method = iter_cb = param_dict['method']
        argsx = (param_dict['y_matrix'][idx], idx, param_dict)
        result.append(minimize(eval_objective, parameters, args=(param_dict['y_matrix'][idx], idx, param_dict),
                 iter_cb=iter_cb, method=method, maxfev=max_iter, nan_policy='omit'))
    return result

def multi_fit(param_dict):
    '''param_dict = {'x_vec': X_vec, 'y_vec': Y_vec, 'z_vec': Z_vec, 'p_vec':P_vec, 'y_matrix': Y_matrix,
                          'ir_x_vec': IR_X_vec, 'ir_y_vec': IR_Y_vec, 'ir_z_vec': IR_Z_vec,
                          'weights_vec': WEIGHTS_vec, 'fxn_vec': FXN_vec, 'fxn_num_vec': FXN_NUM_vec,
                          'param_id_vec': PARAM_ID_vec,'method': method, 'debug': debug, 'group': group, 'cpu': cpu,
                          'ind_fit': ind_fit, 'iter_cb': iter_cb, 'max_iter': max_iter}'''
    idx_list = [*range(len(param_dict['x_vec']))]
    func = partial(optimizer, param_dict)
    cpu_num = int(min(mp.cpu_count()-1, param_dict['cpu'], len(idx_list)))
    results = []
    if cpu_num > 1:
        group = param_dict['group']
        seg_size = int(np.floor(np.floor(len(idx_list) / group) / cpu_num) * group)
        seg_idx_list = [idx_list[i*seg_size:(i+1)*seg_size] if ((i+2)*seg_size) <= len(idx_list) else idx_list[i*seg_size:] for i in range(cpu_num)]
        print(f'Generating Workers (cores:{cpu_num})...')
        proc_pool = mp.Pool(cpu_num)
        print('Fitting data in multiprocessing mode...')
        results = proc_pool.map_async(func, seg_idx_list)
        proc_pool.close()
        proc_pool.join()
        results = list(itertools.chain.from_iterable(results.get()))
    else: # Avoid multiprocessing overhead
        print(f'Fitting data in single core mode...')
        results = [func([idx])[0] for idx in idx_list]
    return results



#
#                         ForteBio Analysis Module Example
#  PyVuka Modules are native python code.  They can import PyVuka as a module to run native
#  PyVuka commands(import ModuleLink.toPyVuka as pyvuka) or access the PyVuka data directly
#  (import data_obj as data).
#-----------------------------------------------------------------------------------------
from .. ModuleLink import toPyVuka as pyvuka
import os
from .. import Modules as IPI
import xml.etree.ElementTree as xmlio
import base64
import pack64
import array
import numpy as np
import json


def auto_analysis():
    pyvuka.new_datamatrix()
    exp_dir = get_exp_dir()
    if not exp_dir:
        return
    readfb(exp_dir)
    sortfb()
    for i in range(1, pyvuka.get_datamatrix().length()+1):
        pyvuka.get_datamatrix().buffer(i).plot.axis.x.lines.show()
    scale_by_delta_load()
    scale_by_mw()
    clip_loading_phase()
    set_association_at_zero()
    buffer_subtract()
    get_amplitudes()
    pyvuka.run_pyvuka_command('fun 40 0')
    guess_initial_parameters()
    set_links()
    fit_all(iter=100)
    apply_amp_cutoff()
    set_y_range()
    calc_score_fxn()
    write_xlsx(os.path.join(os.path.basename(exp_dir + '_processed.xlsx')))
    clean_up(exp_dir)
    pass


def set_association_at_zero():
    origin = pyvuka.get_datamatrix().buffer(1).plot.axis.x.lines.get()[-2]
    origin_pt_num = pyvuka.get_datamatrix().buffer(1).data.x.nearest_index_to_value(origin) + 1
    pyvuka.run_pyvuka_command(f"ori -x -y {1} {pyvuka.get_datamatrix().length()} {origin_pt_num}")


def clip_loading_phase():
    dm = pyvuka.get_datamatrix()
    loading_end_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).plot.axis.x.lines.get()[-3])
    for i in range(1, dm.length() + 1):
        dm.buffer(i).data.x.set(dm.buffer(i).data.x.get()[loading_end_idx:])
        dm.buffer(i).data.y.set(dm.buffer(i).data.y.get()[loading_end_idx:])
        dm.buffer(i).plot.axis.x.lines.set(dm.buffer(i).plot.axis.x.lines.get()[1:])
    pyvuka.set_datamatrix(dm)


def scale_by_mw():
    dm = pyvuka.get_datamatrix()
    mw = 150000
    avg_aa_mw = 120
    Dia_mAb = 12e-9 #12nm length due to biased orientation on sensor
    Dia_ag = 12e-9 #Default value (if no mw supplied) results in no scaling
    assoc_start_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).plot.axis.x.lines.get()[-2])
    for i in range(1, dm.length() +1):
        if 'ag_mw' in dm.buffer(i).meta_dict and float(dm.buffer(i).meta_dict['ag_mw']) > 1:
            mw = float(dm.buffer(i).meta_dict['ag_mw'])
            mw = mw if mw > 1000 else mw*1000
            Dia_ag = np.power((mw/avg_aa_mw), 1/3)*(0.00000000069)
        baseline = dm.buffer(i).data.y.get()[:assoc_start_idx]
        binding = dm.buffer(i).data.y.get()[assoc_start_idx:] * Dia_mAb/Dia_ag
        delta = binding[0] - baseline[-1]
        dm.buffer(i).data.y.set(np.concatenate((baseline, (binding-delta)), axis=None))
        dm.buffer(i).plot.axis.y.title.set("Response Scaled by Loading Amp. & Est. Antigen Diameter\n(Avg Number of Molecules Bound)")
    pyvuka.set_datamatrix(dm)


def scale_by_delta_load():
    dm = pyvuka.get_datamatrix()
    last_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).plot.axis.x.lines.get()[-3])
    new_origin_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).plot.axis.x.lines.get()[-2])
    delta = []
    loading_amp_co = 0.1

    for i in range(1, dm.length() + 1):
        delta.append(dm.buffer(i).data.y.value_at_index(last_idx)-dm.buffer(i).data.y.value_at_index(0))

    mind, maxd = [min(delta), max(delta)]

    t = dm.length()
    s = len(delta)

    for i in range(1, dm.length() + 1):
        dm.buffer(i).data.y.set((dm.buffer(i).data.y.get() - mind) / (maxd-mind))
        if delta[i-1] >= loading_amp_co:
            scalar = 1/abs(dm.buffer(i).data.y.value_at_index(0))
        else:
            scalar = 0
        dm.buffer(i).data.y.set(dm.buffer(i).data.y.get() * scalar)

        # set origin at association and remove loading phase
        cor_x_val = dm.buffer(i).data.x.value_at_index(new_origin_idx)
        cor_y_val = dm.buffer(i).data.y.value_at_index(new_origin_idx)
        dm.buffer(i).data.x.set(dm.buffer(i).data.x.get() - cor_x_val)
        dm.buffer(i).data.y.set(dm.buffer(i).data.y.get() - cor_y_val)
        dm.buffer(i).plot.axis.x.lines.set([x-cor_x_val for x in dm.buffer(i).plot.axis.x.lines.get()])
        dm.buffer(i).plot.axis.y.title.set("Response Scaled by Loading Amp. (nm)")

    pyvuka.set_datamatrix(dm)


def get_exp_dir():
    dir_names = os.listdir(pyvuka.get_current_working_directory())
    dir_names = [os.path.join(pyvuka.get_current_working_directory(), folder) for folder in dir_names if 'processed' != folder.lower() and 'raw' != folder.lower()]
    for name in dir_names:
        exp_dir_list = os.listdir(name)
        lock_list = IPI.get_file_list_nested(name, '.lock')
        file_list = sort_for_exp(IPI.get_file_list_nested(name, '.frd'))
        if not file_list or lock_list or len(exp_dir_list) != 6:
            pass
        else:
            return name
    return False


def readfb(exp_dir):
    file_list = sort_for_exp(IPI.get_file_list_nested(exp_dir, '.frd'))
    for files in file_list:
        ignoreregenerationaandneutralization = True
        Xdata = []
        Ydata = []
        StepName = []
        ActualTime = []
        StepStatus = []
        StepType = []
        Concentration = []
        MolarConcentration = []
        SampleID = []
        WellType = []
        MW = []
        Flags =[]
        SampleGroup = []
        StepLoc = []
        infile = xmlio.parse(os.path.join(exp_dir, files), parser=xmlio.XMLParser(encoding='iso-8859-5')).getroot()
        for expinfo in infile.findall('ExperimentInfo'):
            SensorName = expinfo.find('SensorName').text
            SensorType = expinfo.find('SensorType').text
            SensorRole = expinfo.find('SensorRole').text
            SensorInfo = expinfo.find('SensorInfo').text
        for kindata in infile.findall('KineticsData'):
            for stepdata in kindata.findall('Step'):
                for commondata in stepdata.findall('CommonData'):
                    if ignoreregenerationaandneutralization:
                        WellType.append(commondata.find('WellType').text)
                        if WellType[-1].upper() == 'REGENERATION' or WellType[-1] == 'NEUTRALIZATION':
                            WellType.pop()
                            continue # skip step
                        else:
                            WellType.pop()
                    Concentration.append(commondata.find('Concentration').text)
                    MolarConcentration.append(commondata.find('MolarConcentration').text)
                    SampleID.append(commondata.find('SampleID').text)
                    if commondata.find('SampleGroup') is not None:
                        SampleGroup.append(commondata.find('SampleGroup').text)
                    else:
                        SampleGroup.append(None)
                    WellType.append(commondata.find('WellType').text)
                    MW.append(commondata.find('MolecularWeight').text)
                    Xdata.append(np.array(array.array('f', base64.b64decode(stepdata.find('AssayXData').text))))
                    Ydata.append(np.array(array.array('f', base64.b64decode(stepdata.find('AssayYData').text))))
                    StepName.append(stepdata.find('StepName').text)
                    ActualTime.append(stepdata.find('ActualTime').text)
                    StepStatus.append(stepdata.find('StepStatus').text)
                    StepLoc.append(commondata.find('SampleRow').text + commondata.find('SampleLocation').text)
                    StepType.append(stepdata.find('StepType').text)
        for status in StepStatus:
            if not status == 'OK':
                Flags.append('Sensor:' + status)
                break


        step_initial = 1 # include loading
        step_split = 5
        cycles = 2
        for x in range(0, cycles, 1):
            buffer = pyvuka.new_buffer()
            trimX = Xdata[step_initial:step_split]
            x_to_zero = trimX[0][-1]
            for z, data in enumerate(trimX):
                data = np.array(data)
                if z == 1:
                    data = data - float(x_to_zero + (data[0] - trimX[z-1][-1]))
                elif z > 1:
                    data = data - float(data[0] - buffer.data.x.value_at_index(-1))
                else:
                    data = data-float(x_to_zero)
                buffer.data.x.append(data)
                buffer.plot.axis.x.lines.append(data[0])
            trimY = Ydata[step_initial:step_split]

            for z, data in enumerate(trimY):
                if z == 0:
                    data = data - trimY[0][-1]
                else:
                    data = data-float(data[0]-buffer.data.y.value_at_index(-1))
                buffer.data.y.append(data)

            buffer.data.z.set([float(MolarConcentration[-2])] * buffer.data.y.length())

            buffer.comments.set([SampleID[step_initial] + " on " + SensorType[:3] + " vs " +
                                       SampleID[step_initial+2] + " @ " +
                                       str("{:.2E}".format(float(MolarConcentration[step_initial+2]))) + "nM"])
            buffer.plot.series.name.set(buffer.comments.all_as_string())
            buffer.plot.title.set(buffer.plot.series.name.get())
            buffer.plot.axis.x.title.set("Time (s)")
            buffer.plot.axis.y.title.set("Response (nm)")
            buffer.plot.axis.x.lines.show()
            buffer.meta_dict['ag_mw'] = float(MW[step_initial+2]) if float(MW[step_initial+2]) < 1000 else float(MW[step_initial+2])/1000
            pyvuka.add_buffer_to_datamatrix(buffer)
            step_initial += step_split
            step_split += step_initial
    return "ForteBio data read into memory."


def get_amplitudes():
    dm = pyvuka.get_datamatrix()
    end_base_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).plot.axis.x.lines.get()[-2])
    end_assoc_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).plot.axis.x.lines.get()[-1])
    for i in range(1, dm.length() + 1):
        dm.buffer(i).meta_dict['baseline_amp'] = np.max(dm.buffer(i).data.y.get()[:end_base_idx])
        dm.buffer(i).meta_dict['association_amp'] = np.max(dm.buffer(i).data.y.get()[end_base_idx:end_assoc_idx])
    pyvuka.set_datamatrix(dm)


def sortfb():
    datamatrix_1 = pyvuka.new_datamatrix()
    datamatrix_2 = pyvuka.new_datamatrix()
    dm = pyvuka.get_datamatrix()
    for i in range(1, dm.length() + 1):
        if i%2 != 0:
            datamatrix_1.add_buffer(dm.buffer(i))
        else:
            datamatrix_2.add_buffer(dm.buffer(i))
    pyvuka.set_datamatrix(datamatrix_1)
    pyvuka.append_datamatrix(datamatrix_2)


def sort_for_exp(filelist):
    tuple_list = [(file, int(file.split('_')[-1][:-4])) for file in filelist]
    sorted_tuple_list = sorted(tuple_list, key=lambda x: x[1])
    return_list, key_list = zip(*sorted_tuple_list)
    return return_list


def guess_initial_parameters():
    # reference: Rmax, kd, ka, Cp, m, c, x0, kds = parameters
    dm = pyvuka.get_datamatrix()
    for i in range(1, dm.length() + 1):
        max_y = dm.buffer(i).data.y.max()
        final_y = dm.buffer(i).data.y.value_at_index(-1)
        rmax = max_y
        kd = 0.001
        ka = 60000
        Cp = dm.buffer(i).data.z.value_at_index(0) * 1E-9
        m = 0
        if (final_y <= max_y * 0.2 or final_y >= max_y * 0.8) and final_y >= 0:
            # if really slow off or relatively fast off, assume c = 0.  If final_y < 0, set c
            c = 0
        else:
            # otherwise kinetics are potentially multiphasic or do not otherwise approach 0
            c = final_y

        x0 = dm.buffer(i).plot.axis.x.lines.get()[-1]
        kds = 1
        pyvuka.run_pyvuka_command(f'ap {i} {rmax} {kd} {ka} {Cp} {m} {c} {x0} {kds}')
    return


def buffer_subtract():
    dm = pyvuka.get_datamatrix()
    for i in reversed(range(0, dm.length(), 3)):
        lastbuffer = i+3
        pyvuka.run_pyvuka_command(f'sbf {lastbuffer - 2} {lastbuffer-1} {lastbuffer}')
        pyvuka.run_pyvuka_command(f'sbf {lastbuffer - 2} {lastbuffer - 2} {lastbuffer - 2}')


def set_links():
    # reference: Rmax, kd, ka, Cp, m, c, x0, kds = parameters
    dm = pyvuka.get_datamatrix()
    pyvuka.run_pyvuka_command('fre all')
    pyvuka.run_pyvuka_command('unl all')
    pyvuka.run_pyvuka_command('fix 4')
    pyvuka.run_pyvuka_command('fix 5')
    pyvuka.run_pyvuka_command('fix 6')
    pyvuka.run_pyvuka_command('fix 7')
    pyvuka.run_pyvuka_command('fix 8')
    for i in range(1, dm.length() + 1, 3):
        buffer_set = [i, i+1, i+2]

        # if c!=0, float c
        for j in buffer_set:
            if dm.buffer(j).fit.parameter.get()[5] < 0:
                pyvuka.run_pyvuka_command(f'fix {j} 6')

        # add conditional to not link if both assoc:base amps < 2 (signal cutoff) 12/10/2019
        # ignore buffer_set[0] due to 0nm conc
        assoc_amp_1 = dm.buffer(buffer_set[1]).meta_dict['association_amp']
        base_amp_1 = dm.buffer(buffer_set[1]).meta_dict['baseline_amp']
        assoc_amp_2 = dm.buffer(buffer_set[2]).meta_dict['association_amp']
        base_amp_2 = dm.buffer(buffer_set[2]).meta_dict['baseline_amp']
        if (assoc_amp_2 > base_amp_2 and assoc_amp_1 > base_amp_1) and \
                (assoc_amp_2/base_amp_2 >= 2 and assoc_amp_1/base_amp_1 >= 2):
            pyvuka.run_pyvuka_command(f'lnk {buffer_set[2]} {2} {buffer_set[1]} {2}')
            pyvuka.run_pyvuka_command(f'lnk {buffer_set[2]} {3} {buffer_set[1]} {3}')
    return


def fit_all(iter=np.inf):
    dm = pyvuka.get_datamatrix()
    for i in range(1, len(dm) + 1, 3):
        buffer = dm.buffer(i)
        pyvuka.run_pyvuka_command('pl off')
        pyvuka.run_pyvuka_command(f'pl -v {i} {i+2} {buffer.plot.axis.x.lines.get()[1]} 1E6')
        if np.isinf(iter):
            pyvuka.run_pyvuka_command('fit -debug')
        else:
            pyvuka.run_pyvuka_command(f'fit {int(iter)} -debug')
    pyvuka.run_pyvuka_command('pl off')


def apply_amp_cutoff():
    assoc_amp_co = 0.01
    dm = pyvuka.get_datamatrix()
    for i in range(1, dm.length() + 1, 1):
        if dm.buffer(i).meta_dict['association_amp'] < dm.buffer(i).meta_dict['baseline_amp'] or \
                dm.buffer(i).meta_dict['association_amp'] / dm.buffer(i).meta_dict['baseline_amp'] < 2 or \
                dm.buffer(i).meta_dict['association_amp'] < assoc_amp_co:
            dm.buffer(i).model.x.clear()
            dm.buffer(i).model.y.clear()
            p = dm.buffer(i).fit.parameter.get()
            p[0] = -1
            p[1] = 0
            dm.buffer(i).fit.parameter.set(p)
            dm.buffer(i).fit.parameter_error.set([1E6]*len(dm.buffer(i).fit.parameter_error.get()))
            dm.buffer(i).fit.rsq.set(-1)
        elif dm.buffer(i).model.y.max() > (2 * dm.buffer(i).data.y.max()):
            max_co = 2 * dm.buffer(i).data.y.max()
            y_model = dm.buffer(i).model.y.get()
            # pack64 algo only works with data ranges in certain orders of magnitude.  here we alter unresonable values max co
            dm.buffer(i).model.y.set([y if y <= max_co else max_co for y in y_model])
    pyvuka.set_datamatrix(dm)


def calc_score_fxn():
    dm = pyvuka.get_datamatrix()
    dis_start_idx = dm.buffer(1).data.x.nearest_index_to_value(dm.buffer(1).data.x.get()[-1])
    ideal_stoichiometry = 2
    for i in range(len(dm)):
        first_y = dm.buffer(i).data.y.value_at_index(dis_start_idx)
        perfect_integral = np.abs(first_y * len(dm.buffer(i).data.x.get()[dis_start_idx:]))
        exp_integral = np.sum(dm.buffer(i).data.y.get()[dis_start_idx:])
        score = (exp_integral / perfect_integral) * (first_y / ideal_stoichiometry)
        dm.buffer(i).meta_dict['ipi_score'] = score
    pyvuka.set_datamatrix(dm)


def write_xlsx(xlsx_name):
    output_dir = IPI.create_IPI_output_dir(pyvuka.get_output_directory())

    filename = os.path.join(output_dir, xlsx_name)
    out_file = IPI.xlsx_out(filename)
    out_file.experiment = 'FORTEBIO'
    header = ['Plate', 'Sensor', 'Sensor_Protein', 'Solution_Protein', 'asymtope', 'kd',
              'kd_error', 'ka', 'ka_error', 'Conc', 'Keq', 'Keq_error',
              'Avg_Rsq', 'ipi-score (higher=better)', 'Comments', 'Plot_png', 'JSON']
    out_file.add_header(header)
    out_file.col_widths = [19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 58, 58]
    out_file.row_heights = 228

    dm = pyvuka.get_datamatrix()
    for i in range(1, dm.length() + 1, 3):
        buffer_set = [i, i + 1, i + 2]
        comments = ''
        parameters = ['None' if val is None or np.isnan(val) or not np.isfinite(val) else float(val) for val in
                      dm.buffer(buffer_set[-1]).fit.parameter.get()]
        parameters_err = ['None' if val is None or np.isnan(val) or not np.isfinite(val) else float(val) for val in
                          dm.buffer(buffer_set[-1]).fit.parameter_error.get()]

        Rmax, kd, ka, Cp, m, c, x0, kds = parameters
        if len(parameters_err) != len(parameters):
            parameters_err = [1e6] * len(parameters)
        Rmax_err, kd_err, ka_err, Cp_err, m_err, c_err, x0_err, kds_err = parameters_err

        # catch div by zero
        try:
            KD = kd/ka
        except:
            KD = -1
            comments = 'Fail : KD can not be calculated'

        try:
            KD_err = np.power(np.power(kd_err/kd, 2) + np.power(ka_err/ka, 2), 0.5) * KD
        except:
            KD_err = -1

        # Check rsq values to ensure fit has been made, if fit fails, rsq is nan
        rsq_1 = [-1 if np.isnan(val) or not np.isfinite(val) else float(val) for val in
                 [dm.buffer(buffer_set[0]).fit.rsq.get()]][0]
        rsq_2 = [-1 if np.isnan(val) or not np.isfinite(val) else float(val) for val in
                 [dm.buffer(buffer_set[1]).fit.rsq.get()]][0]
        rsq_3 = [-1 if np.isnan(val) or not np.isfinite(val) else float(val) for val in
                 [dm.buffer(buffer_set[2]).fit.rsq.get()]][0]
        avg_rsq = np.average([x for x in [rsq_1, rsq_2, rsq_3] if x >= 0])
        if np.isnan(avg_rsq):
            avg_rsq = -1
            comments = 'Fail : Insufficient s/n'

        # Check ipi-score values to ensure fit has been made, if fit fails, rsq is nan
        # score_1 = datamatrix[i].item['ipi_score'] 0nm conc
        score_2 = dm.buffer(buffer_set[1]).meta_dict['ipi_score']
        score_3 = dm.buffer(buffer_set[2]).meta_dict['ipi_score']
        avg_score = float(np.average([score_2, score_3]))
        if np.isnan(avg_score) or comments == 'Fail : Insufficient s/n':
            avg_score = -1
            comments = 'Fail : Insufficient s/n'

        # Check ampplitude cutoff of 0.01, fail low amplitudes
        amplitudes = [dm.buffer(buffer_set[0]).meta_dict['association_amp'],
                      dm.buffer(buffer_set[1]).meta_dict['association_amp'],
                      dm.buffer(buffer_set[2]).meta_dict['association_amp']]
        if max(amplitudes) < 0.01:
            comments = 'Fail : Insufficient association amplitude'

        # Check comments for fit failure. Remove model if fit failed
        if 'fail' in comments.lower():
            for j in buffer_set:
                dm.buffer(j).model.x.clear()
                dm.buffer(j).model.y.clear()


        Cp = "{}, {}, {}".format(*[dm.buffer(buffer_set[0]).fit.parameter.get()[3],
                                   dm.buffer(buffer_set[1]).fit.parameter.get()[3],
                                   dm.buffer(buffer_set[2]).fit.parameter.get()[3]])
        sensor_pro = dm.buffer(i).plot.series.name.get().split('on')[0].strip()
        sensor = dm.buffer(i).plot.series.name.get().split('on ')[1].split(' ')[0].strip()
        sol_pro = dm.buffer(i).plot.series.name.get().split('vs')[1].split('@')[0].strip()

        # construct xlsx line entry
        out_line = ['Plate', sensor, sensor_pro, sol_pro, c, kd, kd_err, ka, ka_err, Cp, KD, KD_err,
                    avg_rsq, avg_score, comments, None, None]
        LIMS_dict = {}


        plot_png = pyvuka.get_plot_as_bytestring([i, i+1, i+2], get_bytes=True, black_models=True)

        JSON = {
                'data_x_1': pack64.pack64(dm.buffer(buffer_set[0]).data.x.get()),
                'data_y_1': pack64.pack64(dm.buffer(buffer_set[0]).data.y.get()),
                'model_x_1': pack64.pack64(dm.buffer(buffer_set[0]).model.x.get()),
                'model_y_1': pack64.pack64(dm.buffer(buffer_set[0]).model.y.get()),
                'data_x_2': pack64.pack64(dm.buffer(buffer_set[1]).data.x.get()),
                'data_y_2': pack64.pack64(dm.buffer(buffer_set[1]).data.y.get()),
                'model_x_2': pack64.pack64(dm.buffer(buffer_set[1]).model.x.get()),
                'model_y_2': pack64.pack64(dm.buffer(buffer_set[1]).model.y.get()),
                'data_x_3': pack64.pack64(dm.buffer(buffer_set[2]).data.x.get()),
                'data_y_3': pack64.pack64(dm.buffer(buffer_set[2]).data.y.get()),
                'model_x_3': pack64.pack64(dm.buffer(buffer_set[2]).model.x.get()),
                'model_y_3': pack64.pack64(dm.buffer(buffer_set[2]).model.y.get()),
                'plot_x_lines': str(dm.buffer(buffer_set[0]).plot.axis.x.lines.get())
            }

        out_line[-2] = plot_png
        out_line[-1] = JSON

        for j in range(len(header)):
            key = header[j].lower()
            val = out_line[j]
            if '_png' not in key:
                if key == 'json':
                    key = 'data_xy'
                LIMS_dict.update({key: val})
        out_line[-1] = json.dumps(LIMS_dict)
        out_file.add_line(out_line)

    out_file.write_xlsx()
    return


def set_y_range():
    dm = pyvuka.get_datamatrix()
    y_vals = []
    for i in range(len(dm)):
        buffer = dm.buffer(i)
        y_vals += [buffer.data.y.min(), buffer.data.y.max()]
        if buffer.model.y.length() >= 1:
            y_vals += [buffer.model.y.min(), buffer.model.y.max()]

    ymin, ymax = [min(y_vals)*.95, max(y_vals)*1.05]

    for i in range(1, dm.length() + 1):
        dm.buffer(i).plot.axis.y.range.set([ymin, ymax])
    pyvuka.set_datamatrix(dm)
    return


def clean_up(exp_dir):
    raw_dir = IPI.create_IPI_output_dir(os.path.join(os.path.dirname(exp_dir), 'Raw'))
    IPI.native_copy_nested(exp_dir, os.path.join(raw_dir, os.path.basename(exp_dir)))
    IPI.remove_all_folder(exp_dir)



#!/usr/bin/env python
#title           :ipi_common.py
#author          :R. Paul Nobrega, IPI
#contact         :Paul.Nobrega@ProteinInnovation.com
#description     :This file is a collection of common modules for the Institue
#                 of Protein Innovation custom python scripts. Functions are
#                 specific to data IO and general data handling.
#usage           :Pyvuka module
#python_version  :3.7
#==============================================================================
import datetime
import os
import re
import io
import xlsxwriter as XL
import scipy
import chardet
import matplotlib
matplotlib.use('Agg')
from matplotlib import rcParams
rcParams.update({'figure.autolayout': True})
import matplotlib.pyplot as pl
from matplotlib.ticker import MaxNLocator
import matplotlib.patheffects as pe
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
from io import BytesIO as BIO
from PIL import Image
import shutil
import errno
import multiprocessing
import psutil
import smtplib
from .. import data_obj as data
import stat


##################################################
####  Suportting  Functions Below This Point  ####
##################################################
def local_matrix_to_global_matrix(datamatrix):
    data.matrix = datamatrix
    return True


def find_peaks(one_D_val_list, amp_co, distance_co, rel_peak_co, max_peak_num):
    y_vals = one_D_val_list

    # Get all peaks
    peaks, _ = scipy.signal.find_peaks(y_vals, height=amp_co, distance=distance_co)

    # Sort peaks by y-val
    peak_indx_y = []
    for i in range(len(peaks)):
        peak_indx_y.append([peaks[i], y_vals[peaks[i]]])
    peak_indx_y = sorted(peak_indx_y, key=lambda k: [k[1], k[0]])

    # del items that do not fit rel_peak parameter
    if rel_peak_co is not None:
        max_peak_y = -1E6
        if peak_indx_y:
            max_peak_y = peak_indx_y[-1][1]
        for i in reversed(range(len(peak_indx_y))):
            if (peak_indx_y[i][1] / max_peak_y) * 100 < rel_peak_co:
                del peak_indx_y[i]

    # del peaks that are not local maxima
    idx_shift = int(len(y_vals)*.025)
    for i in reversed(range(len(peak_indx_y))):
        if int(peak_indx_y[i][0]) - idx_shift > 0 and int(peak_indx_y[i][0]) + idx_shift < len(y_vals) - 1:
            trunc_y = y_vals[int(peak_indx_y[i][0]) - idx_shift:int(peak_indx_y[i][0]) + idx_shift]
            if peak_indx_y[i][1] < max(trunc_y):
                del peak_indx_y[i]
        else:
            del peak_indx_y[i]

    # trim to peak counts corresponding to max_peak_num
    peak_indx_y.reverse()
    if len(peak_indx_y) > max_peak_num:
        peak_indx_y = peak_indx_y[:max_peak_num]

    # reconstruct peak list
    peaks = []
    for pair in peak_indx_y:
        peaks.append(pair[0])
    return peaks


def remove_outliers_2d(matrix, SD_CO=2, bins=100, *, slice='xy'):
    '''Assume input xy list constructs 2D XY scatter. This method will bin X and Y values to remove points outside SD limit'''
    # sort incoming list smallest to largest
    y_sorted_matrix = sorted(matrix, key=lambda k: [k[1], k[0]])
    x_sorted_matrix = sorted(matrix, key=lambda k: [k[0], k[1]])
    y_del_map = [False] * len(y_sorted_matrix)
    x_del_map = [False] * len(x_sorted_matrix)
    x_val, y_val = zip(*y_sorted_matrix)
    min_x = min(x_val)
    max_x = max(x_val)
    min_y = min(y_val)
    max_y = max(y_val)
    x_bin_inc = (max_x - min_x + 1)/bins
    y_bin_inc = (max_y - min_y + 1) / bins
    x_bounds = []
    y_bounds = []
    #########################################################
    # Bin by Y and filter out based on x val
    #########################################################
    if 'y' in slice.lower():
        start_index = 0
        stop_index = 0
        for i in range(bins):
            y_band_min = min_y + (y_bin_inc * i)
            y_band_max = y_band_min + y_bin_inc
            for j in range(start_index, len(y_val), 1):
                if y_val[j] <= y_band_min:
                    start_index = j
                elif y_val[j] <= y_band_max:
                    stop_index = j
                else:
                    break
            if start_index < stop_index:
                mean = np.mean(x_val[start_index:stop_index], axis=0)
                sd = np.std(x_val[start_index:stop_index], axis=0)
                avg_y = np.mean(y_val[start_index:stop_index], axis=0)
                x_bounds.append([tuple([mean-(sd*SD_CO), avg_y]), tuple([mean+(sd*SD_CO), avg_y])])
                for k in range(start_index, stop_index, 1):
                    if y_sorted_matrix[k][0] < mean - SD_CO * sd or y_sorted_matrix[k][0] > mean + SD_CO * sd:
                        y_del_map[k] = True
            start_index = stop_index

    #########################################################
    # Bin by X and filter out based on y val
    #########################################################
    if 'x' in slice.lower():
        x_val, y_val = zip(*x_sorted_matrix)
        start_index = 0
        stop_index = 0
        for i in range(bins):
            x_band_min = min_x + (x_bin_inc * i)
            x_band_max = x_band_min + x_bin_inc
            for j in range(start_index, len(x_val), 1):
                if x_val[j] <= x_band_min:
                    start_index = j
                elif x_val[j] <= x_band_max:
                    stop_index = j
                else:
                    break
            if start_index < stop_index:
                mean = np.mean(y_val[start_index:stop_index], axis=0)
                sd = np.std(y_val[start_index:stop_index], axis=0)
                avg_x = np.mean(x_val[start_index:stop_index], axis=0)
                y_bounds.append([tuple([avg_x, mean-(sd*SD_CO)]), tuple([avg_x, mean+(sd*SD_CO)])])
                for k in range(start_index, stop_index, 1):
                    if x_sorted_matrix[k][1] < mean - SD_CO * sd or x_sorted_matrix[k][1] > mean + SD_CO * sd:
                        x_del_map[k] = True
            start_index = stop_index

    #########################################################
    # Apply del map to sorted_matrix
    #########################################################
    clean_matrix = []
    for i in range(len(y_sorted_matrix)):
        if not y_del_map[i]:
            clean_matrix.append(y_sorted_matrix[i])
    del_clean_matrix = [False] * len(clean_matrix)
    for i in range(len(x_sorted_matrix)):
        if x_del_map[i] and x_sorted_matrix[i] in clean_matrix:
            del_clean_matrix[clean_matrix.index(x_sorted_matrix[i])] = True
    for i in reversed(range(len(clean_matrix))):
        if del_clean_matrix[i]:
            del clean_matrix[i]

    return clean_matrix, x_bounds, y_bounds


def calc_integral(buffer, start_idx, stop_idx):
    # not the best way to calculate this, but output matches AKTA software
    integral = 0.000
    min_y = min(buffer.item['y_val'])
    x_vals = np.array(buffer.item['x_val'][start_idx:stop_idx])
    y_vals = np.array(buffer.item['y_val'][start_idx:stop_idx]) - min_y
    temp = np.average(np.diff(x_vals))
    integral = sum(y_vals * temp)
    return integral


def calc_linear_fit(x_list, y_list):
    """ Computes the least-squares solution to a linear matrix equation. """
    N = len(x_list)
    x_avg = sum(x_list) / N
    y_avg = sum(y_list) / N
    var_x, cov_xy = 0, 0
    for x, y in zip(x_list, y_list):
        temp = x - x_avg
        var_x += temp ** 2
        cov_xy += temp * (y - y_avg)
    slope = cov_xy / var_x
    y_interc = y_avg - slope * x_avg
    return (slope, y_interc)


def predict_encoding(file_path, n_lines=20):
    '''Predict a file's encoding using chardet'''
    # Open the file as binary data
    with open(file_path, 'rb') as f:
        # Join binary lines for specified number of lines
        rawdata = b''.join([f.readline() for _ in range(n_lines)])

    return chardet.detect(rawdata)['encoding']


def txt_to_list(filename):
    # Read in text file
    in_txt = []
    encode_type = predict_encoding(filename)
    delim = detect_delimiter(filename)
    with io.open(filename, 'r', encoding=encode_type) as textFile:
        for line in textFile:
            in_txt.append(line.split(delim))
    return in_txt


def detect_delimiter(filename):
    '''Determine if comma or tab delimited'''
    encode_type = predict_encoding(filename)
    with io.open(filename, 'r', encoding=encode_type) as textFile:
        linein = textFile.readline()
        if len(linein.split(',')) < len(linein.split('\t')):
            return '\t'
        else:
            return ','


def is_float(input):
    try:
        num = float(input)
    except ValueError:
        return False
    return True


def is_int(input):
    try:
        num = int(input)
    except ValueError:
        return False
    return True


def get_file_list(folder, ext):
    '''Returns filelist of all files within a folder having a prescribed extension

    Parameters
    ----------
    folder: types.StringType object
        Description of 'folder'. Valid OS path of directory containing files to construct a list of
    ext: types.StringType object
        Description of 'ext'. File extension to filter file list by

    Returns
    ----------
    types.ListType object of types.StringType objects
        Returns list of valid OS path filenames
    '''
    file_list = list(filter(lambda x: x.lower().endswith(ext), os.listdir(folder)))
    # Make full path and sort for most recent modification date first, and return
    return sorted([os.path.join(folder, f) for f in file_list], key=(lambda x: os.path.getmtime(x))) if len(file_list) > 0 else []


def get_file_list_nested(folder, ext):
    '''Returns filelist of all files within a folder and subfolders when having a prescribed extension

    Parameters
    ----------
    folder: types.StringType object
        Description of 'folder'. Valid OS path of directory containing files to construct a list of
    ext: types.StringType object
        Description of 'ext'. File extension to filter file list by

    Returns
    ----------
    types.ListType object of types.StringType objects
        Returns list of valid OS path filenames
    '''

    listOfFiles = list()
    for (dirpath, dirnames, filenames) in os.walk(folder):
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]

    return_list = sorted(list(filter(lambda x: x.lower().endswith(ext), listOfFiles)))

    return return_list if len(return_list) > 0 else []


def change_file_ext(filename, new_ext):
    if new_ext[0] != '.':
        new_ext = '.' + new_ext
    base, _ = os.path.splitext(filename)
    # if file already exists, make change
    if os.path.isfile(filename):
        os.rename(filename, base + new_ext)
    # return new filename
    return base + new_ext


'''def find_nearest(array, value):
    array = np.asarray(array)
    subarray = np.abs(array - value)
    idx = subarray.argmin()
    return idx'''


def colorallseries(datamatrix):
    start = 0.0
    stop = 1.0
    number_of_buffers = len(datamatrix)
    cm_subsection = np.linspace(start, stop, number_of_buffers)

    colors = [pl.cm.gist_rainbow(x) for x in cm_subsection]

    for i, color in enumerate(colors):
        datamatrix[i].item['series_color'] = color
    return datamatrix


def seriestotitle(datamatrix):
    for i in range(1, len(datamatrix) + 1):
        datamatrix.buffer(i).plot.title.set(datamatrix.buffer(i).plot.series.name.get())
    return datamatrix


def native_extract_nested_zip(zippedFile, toFolder):
    """ Extract a zip file including any nested zip files
        Delete the zip file(s) after extraction
        7zip used for windows and unzip for unix.

        System level operations are waaayyy faster than zipfile
    """
    command = 'unzip -n \"{}\" -d \"{}\"'.format(*[zippedFile, toFolder])
    if os.name == 'nt':
        if os.path.isfile(r"C:\Program Files\7-Zip\7z.exe"):
            zipexc = r'C:\PROGRA~1\7-Zip\7z.exe'
        else:
            zipexc = '7z.exe'
        command = '{} x \"{}\" -o\"{}\" -aos'.format(*[zipexc, zippedFile, toFolder])
    os.system(command)
    #os.remove(zippedFile)
    for root, dirs, files in os.walk(toFolder):
        for filename in files:
            if re.search(r'\.zip$', filename):
                fileSpec = os.path.join(root, filename)
                command = 'unzip -n \"{}\" -d \"{}\"'.format(*[fileSpec, root])
                if os.name == 'nt':
                    if os.path.isfile(r"C:\Program Files\7-Zip\7z.exe"):
                        zipexc = r'C:\PROGRA~1\7-Zip\7z.exe'
                    else:
                        zipexc = '7z.exe'
                    command = '{} x \"{}\" -o\"{}\" -aos'.format(*[zipexc, fileSpec, root])
                os.system(command)
                #os.remove(fileSpec)
                for new_file in os.listdir(root):
                    if '.zip' not in new_file:
                        file_out = os.path.basename(root) + '!' + new_file
                        if not os.path.isfile(os.path.join(toFolder, file_out)):
                            shutil.copy(os.path.join(root, new_file), os.path.join(toFolder, file_out))
    return True


def native_copy(src, dest):
    """For Large Files, This is waaayyyy faster than shutil"""
    command = 'copy' if os.name == 'nt' else 'cp'
    os.system('{} \"{}\" \"{}\"'.format(*[command, src, dest]))
    return True


def native_copy_nested(src, dest):
    """For Large File trees, This is waaayyyy faster than shutil"""
    if os.name == 'nt':
        os.system('Xcopy /E /I /Y \"{}\" \"{}\"'.format(*[src, dest]))
    else:
        os.system('yes | cp -a \"{}.\" \"{}\"'.format(*[src, dest]))
    return True


def remove_all_folder(folder):
    def on_rm_error(func, path, exc_info):
        # path contains the path of the file that couldn't be removed
        # let's just assume that it's read-only and unlink it.
        os.chmod(path, stat.S_IWRITE)
        os.unlink(path)

    shutil.rmtree(folder, onerror=on_rm_error)
    return True


def move_file(src, dest):
    return shutil.move(src, dest)


def move_entire_directory(src, dest):
    try:
        shutil.copytree(src, dest)
    except OSError as e:
        # If the error was caused because the source wasn't a directory
        if e.errno == errno.ENOTDIR:
            shutil.copy(src, dest)
        else:
            return 'ERROR: Directory not copied: %s' % str(e)
    shutil.rmtree(src, True, None)
    return dest


def create_IPI_output_dir(parent_output_dir):
    today = datetime.date.today()
    month = today.strftime('%m')
    day = today.strftime('%d')
    year = today.strftime('%Y')
    year_dir = os.path.join(parent_output_dir, year)
    month_dir = os.path.join(year_dir,month)
    day_dir = os.path.join(month_dir, day)
    dir_hierarchy = [parent_output_dir, year_dir, month_dir, day_dir]
    return list(map(lambda x: check_directory(x), dir_hierarchy))[-1]


def check_directory(dir_name):
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    return dir_name


def calc_derivative(x_list, y_list):
    x_out = []
    y_out = []
    for i in range(len(x_list)):
        if i == 0:
            x_out.append(x_list[i])
            y_out.append((y_list[i+1] - y_list[i]) / (x_list[i+1] - x_list[i]))
        else:
            x_out.append(x_list[i])
            y_out.append((y_list[i] - y_list[i-1]) / (x_list[i] - x_list[i-1]))
    return x_out, y_out


def safe_execute(default, exception, function, *args):
    '''
    Executes passed function in try / except
    Borrowed from: https://stackoverflow.com/questions/36671077/one-line-exception-handling/36671208
    :param default: Value to return upon exception
    :param exception: Specific Exception to catch
    :param function: Function to evaluate within try
    :param args: Arguments for function being passed
    :return: Result of successful execution of function or 'default' upon Exception
    '''
    try:
        return function(*args)
    except exception:
        return default
##################################################
####      Object Classes Below This Point     ####
##################################################
class buffer(object):
    def __init__(self):
        self.item = {}
        self.item['x_val'] = []
        self.item['xe_val'] = []
        self.item['y_val'] = []
        self.item['ye_val'] = []
        self.item['z_val'] = []
        self.item['ze_val'] = []
        self.item['x_category'] = []
        self.item['y_category'] = []
        self.item['x_model'] = []
        self.item['y_model'] = []
        self.item['z_model'] = []
        self.item['x_residuals'] = []
        self.item['y_residuals'] = []
        self.item['z_residuals'] = []
        self.item['instrument_response_x'] = []
        self.item['instrument_response_y'] = []
        self.item['instrument_response_z'] = []
        self.item['fit_fxn'] = []
        self.item['fit_fxn_index'] = []
        self.item['parameters'] = []
        self.item['parameters_error'] = []
        self.item['fit_xsq'] = 0
        self.item['fit_rsq'] = 0
        self.item['link'] = []
        self.item['free'] = []
        self.item['comments'] = []
        self.item['series_name'] = ''
        self.item['series_color'] = ''
        self.item['series_type'] = ''
        self.item['series_weight'] = 10
        self.item['plot_stack'] = []
        self.item['data_point_range'] = []
        self.item['plot_type'] = ""
        self.item['plot_title'] = ""
        self.item['plot_x_axis_title'] = ""
        self.item['plot_y_axis_title'] = ""
        self.item['plot_z_axis_title'] = ""
        self.item['plot_x_axis_type'] = "linear"
        self.item['plot_y_axis_type'] = "linear"
        self.item['plot_z_axis_type'] = "linear"
        self.item['plot_x_axis_range'] = []
        self.item['plot_y_axis_range'] = []
        self.item['plot_z_axis_range'] = []
        self.item['plot_y_lines'] = []
        self.item['plot_x_lines'] = []
        self.item['plot_polygons'] = []
        self.item['plot_peak_bounds'] = []
        self.item['plot_peaks'] = []
        self.item['plot_integrals'] = []
        self.item['plot_show_y_lines'] = False
        self.item['plot_show_x_lines'] = False
        self.item['plot_show_polygons'] = False
        self.item['plot_show_peak_bounds'] = False
        self.item['plot_show_peaks'] = False
        self.item['plot_show_integrals'] = False
        self.item['plot_show_unity'] = False
        self.item['plot_show_x_error_bars'] = False
        self.item['plot_show_y_error_bars'] = False
        self.item['plot_show_z_error_bars'] = False
        self.item['plot_show_integrals'] = False
        self.item['plot_show_peak_bounds'] = False
        self.item['plot_show_peaks'] = False
        self.item['plot_show_legend'] = False
        self.item['plot_unweighted_residuals'] = False


class xlsx_out(object):

    def __init__(self, filename):
        """Return an xlsx object whose name is filename """
        self.filename = filename
        self.matrix = []
        self.header = []
        self.col_widths = []
        self.row_heights = []
        self.sheet_name = 'output'
        self.lims_upload = True
        self.manual_review = False
        self.experiment = 'None'

    def add_line(self, write_array):
        self.matrix.append(write_array)
        return

    def add_header(self, write_array):
        self.header = write_array
        return

    def set_col_widths(self, widths):
        if not widths:
            return
        maxcol = -10
        if ',' in str(widths):
            array_out = list(widths)
        else:
            if not self.header:
                self.col_widths = widths
                return
            if len(self.header) > maxcol:
                maxcol = len(self.header)
            for line in self.matrix:
                if len(line) > maxcol:
                    maxcol = len(line)
            array_out = [widths for j in range(maxcol)]
        self.col_widths = array_out
        return

    def set_row_heights(self, heights):
        if not self.header:
            header = 0
        else:
            header = 1
        if not heights:
            return
        maxrow = -10
        if ',' in str(heights):
            array_out = list(heights)
        else:
            if not self.matrix:
                self.row_heights = heights
                return
            elif len(self.matrix) > maxrow:
                maxrow = len(self.matrix)
            array_out = [heights for j in range(maxrow + header)]
        self.row_heights = array_out
        return

    def __copy_to_lims_dir(self):
        output_full_name = self.filename
        file_name = os.path.basename(self.filename)
        main_dir = os.path.abspath(os.path.splitdrive(output_full_name)[0])
        upload_dir_file = os.path.join(main_dir, 'lims_upload', file_name)
        print("\nCopying processed data to LIMS dir: " + upload_dir_file)
        print(output_full_name + ' --> ' + os.path.splitext(upload_dir_file)[0] + '_' + self.experiment.upper() + os.path.splitext(upload_dir_file)[1])
        shutil.copy(output_full_name, os.path.splitext(upload_dir_file)[0] + '_' +
                    self.experiment.upper() + os.path.splitext(upload_dir_file)[1])

    def write_xlsx(self):
        self.set_row_heights(self.row_heights)
        self.set_col_widths(self.col_widths)
        wb = XL.Workbook(self.filename)
        ws = []
        ws.append(wb.add_worksheet(self.sheet_name))
        header = 0
        fail_format = wb.add_format({'bold': True, 'bg_color': '#FFC7CE'})
        pass_format = wb.add_format({'bold': True, 'bg_color': '#C6EFCE'})

        if self.header:
            for i in range(len(self.header)):
                ws[0].write(0, i, self.header[i])
            ws[0].freeze_panes(1, 0)
            header = 1

        if self.col_widths:
            if not isinstance(self.col_widths, list):
                self.col_widths = [float(self.col_widths)] * int(len(self.header))
            for i in range(len(self.col_widths)):
                ws[0].set_column(i, i, self.col_widths[i])

        if self.row_heights:
            if not isinstance(self.row_heights, list):
                self.row_heights = [float(self.row_heights)] * int(len(self.header))
            for i in range(len(self.row_heights)):
                ws[0].set_row(i, self.row_heights[i])
            if self.header:
                ws[0].set_row(0, 18)

        for i in range(len(self.matrix)):
            for j in range(len(self.matrix[i])):
                var_type = str(type(self.matrix[i][j])).lower()
                var_to_str = str(self.matrix[i][j]).lower()
                if 'bytesio' in var_type:
                    # if image, scale to cell height
                    img_out = Image.open(self.matrix[i][j])
                    img_width, img_height = img_out.size
                    # 1.33 is used to correct xlsx cell height to px
                    scalar = (self.row_heights[i+header] * 1.33)/img_height
                    # insert image, with scaling
                    ws[0].insert_image(i+header, j, 'figure.png',
                                   {'image_data': self.matrix[i][j], 'x_scale': scalar, 'y_scale': scalar})
                elif var_to_str in ['true', 'pass'] or 'true' in var_to_str:
                    ws[0].write(i + header, j, self.matrix[i][j], pass_format)
                elif var_to_str in ['false', 'fail'] or 'fail' in var_to_str:
                    ws[0].write(i + header, j, self.matrix[i][j], fail_format)
                elif var_to_str in ['inf', '-inf', 'nan'] and ('float' in var_type or 'int' in var_type):
                    ws[0].write(i + header, j, -1)
                else:
                    ws[0].write(i+header, j, self.matrix[i][j])
        wb.close()
        if self.lims_upload:
            self.__copy_to_lims_dir()
        else:
            new_name = os.path.splitext(self.filename)[0] + '_' + self.experiment + os.path.splitext(self.filename)[1]
            if os.path.isfile(new_name):
                os.remove(new_name)
            os.rename(self.filename, new_name)
            if self.manual_review:
                email = emailer()
                email.send_to_all('ATTN: Review Data',
                                  'The following data requires manual review before upload to the LIMS. '
                                  'Please edit the XLSX file accordingly, then copy the file to the '
                                  '\'lims_upload\' directory within the'
                                  ' Automation folder.\n\nFile:\t{}'.format(new_name))
        return


class plotter(object):
    def __init__(self):
        pl.close('all')
        print("Plotting...")

    def __call__(self, buffer, *args, **kwargs):
        return_bytes = False
        dpi = 50

        if 'get_bytes' in kwargs:
            return_bytes = bool(kwargs['get_bytes'])
        if 'dpi' in kwargs:
            dpi = int(kwargs['dpi'])

        plot_type = self.__get_plot_type(buffer)

        if plot_type == 'heatmap':
            pl = self.__heatmap(buffer, *args, **kwargs)
        else:
            pl = self.__scatterline(buffer, plot_type, *args, **kwargs)

        if return_bytes:
            return self.__get_byte_png(pl, dpi=dpi)
        return "Plotting Complete"

    def __get_plot_type(self, buffer):
        all_plots = ['scatter', 'line', 'heatmap']
        default = 'scatter'

        plot_type = buffer.item['plot_type']

        if plot_type.lower() in all_plots:
            return plot_type.lower()

        X = np.array(buffer.item['x_val'])
        XE = np.array(buffer.item['xe_val'])
        Y = np.array(buffer.item['y_val'])
        YE = np.array(buffer.item['ye_val'])
        Z = np.array(buffer.item['z_val'])
        ZE = np.array(buffer.item['ze_val'])
        X_model = np.array(buffer.item['x_model'])
        Y_model = np.array(buffer.item['y_model'])
        Z_model = np.array(buffer.item['z_model'])
        X_cat = buffer.item['x_category']
        Y_cat = buffer.item['y_category']

        if len(X) > 1 and len(Y) > 1 and len(Z) == 0 and (len(X_model) == 0 or len(Y_model) == 0):
            return 'line'
        elif len(X_cat) > 1 and len(Y_cat) > 1 and len(Z) > 1:
            return 'heatmap'
        return default

    def __heatmap(self, buffer, *args, **kwargs):
        X = np.array(buffer.item['x_val'])
        Y = np.array(buffer.item['y_val'])
        Z = np.array(buffer.item['z_val'])
        X_cat = buffer.item['x_category']
        Y_cat = buffer.item['y_category']
        X_title = buffer.item['plot_x_axis_title']
        Y_title = buffer.item['plot_y_axis_title']
        Z_title = buffer.item['plot_z_axis_title']
        plot_title = buffer.item['plot_title']

        X_axis = X if not X_cat else X_cat
        Y_axis = Y if not Y_cat else Y_cat

        data = Z.reshape((len(Y_axis), len(X_axis)))

        data = np.array(data)

        fig, ax = pl.subplots()
        pl.rcParams["font.size"] = 10
        pl.gca().tick_params(axis='y', pad=8)
        pl.gca().tick_params(axis='x', pad=8)

        # Plot the heatmap
        im = ax.imshow(data, cmap=matplotlib.cm.rainbow)

        # We want to show all ticks...
        ax.set_xticks(np.arange(data.shape[1]))
        ax.set_yticks(np.arange(data.shape[0]))
        # ... and label them with the respective list entries.
        ax.set_xticklabels(X_axis)
        ax.set_yticklabels(Y_axis)
        pl.xlabel(X_title)
        pl.ylabel(Y_title)

        # Let the horizontal axes labeling appear on top.
        ax.tick_params(top=True, bottom=False,
                       labeltop=True, labelbottom=False)

        # Rotate the tick labels and set their alignment.
        pl.setp(ax.get_xticklabels(), rotation=-30, ha="right",
                 rotation_mode="anchor")

        # Turn spines off and create white grid.
        for edge, spine in ax.spines.items():
            spine.set_visible(False)
        ax.set_xticks(np.arange(data.shape[1] + 1) - .5, minor=True)
        ax.set_yticks(np.arange(data.shape[0] + 1) - .5, minor=True)
        ax.grid(which="minor", color="w", linestyle='-', linewidth=2)
        ax.tick_params(which="minor", bottom=False, left=False)

        # Loop over data dimensions and create text annotations.
        for i in range(len(Y_axis)):
            for j in range(len(X_axis)):
                ax.text(j, i, data[i, j], ha="center", va="center", color='k')

        if len(plot_title) > 1:
            ax.set_title(plot_title, pad=30, loc='center')

        # Create colorbar
        # create an axes on the right side of ax. The width of cax will be 5%
        # of ax and the padding between cax and ax will be fixed at 0.1 inch.
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)
        cbar = pl.colorbar(im, cax=cax)
        cbar.ax.set_ylabel(Z_title, rotation=-90, va="bottom", labelpad=5)

        fig.canvas.flush_events()
        return pl

    def __scatterline(self, buffer, plot_type, *args, **kwargs):
        return_bytes = False
        black_models = False

        if 'get_bytes' in kwargs:
            return_bytes = bool(kwargs['get_bytes'])
        if 'black_models' in kwargs:
            black_models = kwargs['black_models']

        # Plot Data-model
        xmin = min(buffer.item['x_val'])
        xmax = max(buffer.item['x_val'])

        #####################################
        # declare variables to tidy up code #
        # ___________________________________#
        X = np.array(buffer.item['x_val'])
        XE = np.array(buffer.item['xe_val'])
        Y = np.array(buffer.item['y_val'])
        YE = np.array(buffer.item['ye_val'])
        Z = np.array(buffer.item['z_val'])
        ZE = np.array(buffer.item['ze_val'])
        X_model = np.array(buffer.item['x_model'])
        Y_model = np.array(buffer.item['y_model'])
        Z_model = np.array(buffer.item['z_model'])
        X_resid = np.array(buffer.item['x_residuals'])
        Y_resid = np.array(buffer.item['y_residuals'])
        Z_resid = np.array(buffer.item['z_residuals'])
        series_color = [val if val != '' else 'red' for val in [buffer.item['series_color']]][0]
        series_weight = float(buffer.item['series_weight'])
        x_axis_title = str(buffer.item['plot_x_axis_title'])
        y_axis_title = str(buffer.item['plot_y_axis_title'])
        z_axis_title = str(buffer.item['plot_z_axis_title'])
        show_x_lines = bool(buffer.item['plot_show_x_lines'])
        x_lines = buffer.item['plot_x_lines']
        show_y_lines = bool(buffer.item['plot_show_y_lines'])
        y_lines = buffer.item['plot_y_lines']
        show_peaks = bool(buffer.item['plot_show_peaks'])
        peak_indicies = buffer.item['plot_peaks']
        show_peak_bounds = buffer.item['plot_show_peak_bounds']
        peak_bounds_indicies = buffer.item['plot_peak_bounds']
        show_integrals = bool(buffer.item['plot_show_integrals'])
        integral_indicies_pairs = buffer.item['plot_integrals']
        show_polygons = bool(buffer.item['plot_show_polygons'])
        polygon_verticies = buffer.item['plot_polygons']
        x_axis_type = str(buffer.item['plot_x_axis_type'])
        y_axis_type = str(buffer.item['plot_y_axis_type'])
        z_axis_type = str(buffer.item['plot_z_axis_type'])
        x_axis_range = buffer.item['plot_x_axis_range']
        y_axis_range = buffer.item['plot_y_axis_range']
        z_axis_range = buffer.item['plot_z_axis_range']
        # ___________________________________#
        # END Local Variable Definitions    #
        #####################################

        fig, ax = pl.subplots(nrows=1, ncols=1, figsize=(8, 6), sharey=True, sharex=True)
        ax.plot(1, 1)
        ax.set_xscale(x_axis_type)
        ax.set_yscale(y_axis_type)
        pl.rcParams["font.size"] = 10
        pl.gca().tick_params(axis='y', pad=8)
        pl.gca().tick_params(axis='x', pad=8)

        if len(x_axis_range) > 0:
            # do nothing
            pass
        else:
            x_axis_range = (min(X), max(X))

        if len(y_axis_range) > 0:
            # do nothing
            pass
        else:
            padding = (max(Y) - min(Y)) * 0.05
            y_axis_range = (min(Y) - padding, max(Y) + padding)

        # Set Plot Labels
        pl.xlim(min(x_axis_range), max(x_axis_range))
        pl.ylim(min(y_axis_range), max(y_axis_range))
        pl.xlabel(x_axis_title)
        pl.ylabel(y_axis_title)

        # If Z data, treat as heatmap scatter
        if len(Z) == 0:
            if plot_type == 'scatter':
                ax.scatter(X, Y, color=series_color, s=series_weight * 4, zorder=0)  # weight is by area, correct to height val
            else:
                ax.plot(X, Y, color=series_color, linestyle='-', linewidth=series_weight, zorder=0)
        else:
            # color points by z value and sort to not obscure data in plot
            sorted_X, sorted_Y, sorted_Z = self.__sort_by_z(X, Y, Z)
            if plot_type == 'scatter':
                ax.scatter(sorted_X, sorted_Y, c=sorted_Z, s=series_weight * 4, cmap=matplotlib.cm.rainbow, zorder=0)  # weight is by area, correct to height val
            else:
                ax.plot(sorted_X, sorted_Y, c=sorted_Z, linestyle='-', linewidth=series_weight, cmap=matplotlib.cm.rainbow, zorder=0)

        #####################################
        #            Residual Plot          #
        # ___________________________________#
        if len(X_model) > 0:
            if black_models:
                ax.plot(X_model, Y_model, linestyle='-', color='k', linewidth=series_weight / 5, zorder=1)
            else:
                ax.plot(X_model, Y_model, linestyle='-', color=series_color, linewidth=series_weight / 5,
                        path_effects=[pe.Stroke(linewidth=series_weight / 2.5, foreground='k'), pe.Normal()],
                        zorder=1)
            # residuals plot
            divider = make_axes_locatable(ax)
            ax2 = divider.append_axes("bottom", size="20%", pad=0)
            ax.figure.add_axes(ax2)
            ax2.set_xscale(x_axis_type)
            ax2.set_yscale(y_axis_type)
            # end residuals plot
            ax.set_xticklabels([])
            pl.gca().yaxis.set_major_locator(MaxNLocator(prune='upper'))
            # Set x bounds to match data plot
            pl.xlim(min(x_axis_range), max(x_axis_range))
            pl.xlabel(x_axis_title)
            ax2.plot(X, Y_resid, linestyle='-', color=series_color, linewidth=series_weight / 5)
            ax2.axhline(y=0, linestyle='-', color='k', linewidth=2, zorder=0)
            ax2.grid()
        # ___________________________________#
        #        END Residual Plot          #
        #####################################

        #####################################
        #     X-line and Y-line Plotting    #
        # ___________________________________#
        if show_x_lines and len(x_lines) >= 1:
            for line in x_lines:
                ax.axvline(x=line, linestyle='--', color='gray', linewidth=series_weight / 5, zorder=2,
                           path_effects=[pe.Stroke(linewidth=series_weight / 2.5, foreground='k'), pe.Normal()])
        if show_y_lines and len(y_lines) >= 1:
            for line in y_lines:
                ax.axhline(y=line, linestyle='--', color='gray', linewidth=series_weight / 5, zorder=2,
                           path_effects=[pe.Stroke(linewidth=series_weight / 2.5, foreground='k'), pe.Normal()])
        # ___________________________________#
        #  END X-line and Y-line Plotting   #
        #####################################

        #####################################
        #           Peak Plotting           #
        # ___________________________________#
        if show_peaks and len(peak_indicies) >= 1:
            peak_x = []
            peak_y = []
            for p in peak_indicies:
                peak_x.append(X[p])
                peak_y.append(Y[p])
            ax.plot(peak_x, peak_y, "d", markerfacecolor='gainsboro', markeredgecolor='k', zorder=3)
        if show_peak_bounds and len(peak_bounds_indicies) >= 1:
            peak_bound_x = []
            peak_bound_y = []
            for p in peak_bounds_indicies:
                p_start = p[0]
                p_end = p[1]
                peak_bound_x.append(X[p_start])
                peak_bound_y.append(Y[p_start])
                peak_bound_x.append(X[p_end])
                peak_bound_y.append(Y[p_end])
            ax.plot(peak_bound_x, peak_bound_y, "d", markerfacecolor='gainsboro', markeredgecolor='k', zorder=3)
        # ___________________________________#
        #       END Peak Plotting           #
        #####################################

        #####################################
        #         Integral  Plotting        #
        # ___________________________________#
        if show_integrals and len(integral_indicies_pairs) >= 1:
            if not isinstance(series_color, tuple):
                series_color = matplotlib.colors.to_rgba(series_color)
            int_color = list(series_color)
            int_color[3] = 0.4
            int_color = tuple(int_color)
            int_x = []
            int_y = []
            xy_vals = list(zip(X, Y))
            for int_indx in integral_indicies_pairs:
                int_start = (X[int_indx[0]], Y[int_indx[0]])
                int_end = (X[int_indx[-1]], Y[int_indx[-1]])
                verts = [(int_start[0], 0)] + xy_vals[int_indx[0]:int_indx[1]] + [(int_end[0], 0)]
                poly = matplotlib.patches.Polygon(verts, facecolor=int_color, edgecolor='gainsboro', zorder=0)
                ax.add_patch(poly)
        # ___________________________________#
        #     END  Integral  Plotting       #
        #####################################

        #####################################
        #        Polygon Plotting           #
        # ___________________________________#
        if show_polygons and len(polygon_verticies) >= 1:
            for polygon in polygon_verticies:
                poly_x, poly_y = zip(*polygon)
                ax.plot(poly_x, poly_y, linestyle='-', color='gainsboro', linewidth=series_weight / 5, zorder=4,
                        path_effects=[pe.Stroke(linewidth=series_weight / 2.5, foreground='k'), pe.Normal()])
        # ___________________________________#
        #       END  Polygon Plotting       #
        #####################################

        ax.grid()
        fig.canvas.flush_events()

        return pl

    def __get_byte_png(self, pl, dpi=50):
        BIOstream = BIO()
        pl.tight_layout()
        pl.savefig(BIOstream, format='png', dpi=dpi, bbox_inches='tight')
        BIOstream.seek(0)
        im = Image.open(BIOstream)
        im2 = im.convert('RGB').convert('P', palette=Image.ADAPTIVE)
        im2 = im2.resize(im2.size, Image.ANTIALIAS)
        BIOstream = BIO()
        im2.save(BIOstream, format='PNG', quality=95)
        BIOstream.seek(0)
        byte_png = BIOstream.getvalue()
        pl.close()
        return byte_png

    def __sort_by_z(self, X, Y, Z):
        idx = Z.argsort()
        sorted_x, sorted_y, sorted_z = X[idx], Y[idx], Z[idx]
        return sorted_x, sorted_y, sorted_z


class multiproc(object):
    def __init__(self, target_def):
        """Run a method on all available cpus"""
        self.__target = target_def
        self.args = []
        self.__cpu_num = psutil.cpu_count(logical=True)
        self.result = []

    def add_args(self, arg_lst):
        self.args.append(arg_lst)
        return

    def set_cpu_num(self, cpu):
        self.__cpu_num = int(cpu)
        return

    def start(self):
        number_of_processes = self.__cpu_num
        arg_set = [tuple(a) for a in self.args]
        with multiprocessing.Pool(processes=number_of_processes) as pool:
            results = pool.starmap(self.__target, arg_set)
        if isinstance(results, list) and isinstance(results[0], list):
            for item in results:
                self.result.extend(item)
        else:
            self.result = results


class emailer(object):
    def __init__(self):
        #common e-mail related parameters
        self.__admin_addr = 'paul.nobrega@proteininnovation.org'
        self.__all_addr = 'paul.nobrega@proteininnovation.org' #'lab@proteininnovation.org'
        self.__from_addr = 'Automation@ProteinInnovation.org'
        self.__from_pwd = '3qPKdUcEtU'
        self.__smtp_port = 587
        self.__smtp_host = 'smtp.gmail.com'
        self.__sys_name = 'IPI_Automation'

    def __send_email(self, message, to_user):
        try:
            print(f"Sending e-mail to {to_user}...")
            server = smtplib.SMTP(self.__smtp_host, self.__smtp_port)
            server.ehlo()
            server.starttls()
            server.login(self.__from_addr, self.__from_pwd)
            server.sendmail(self.__from_addr, self.__all_addr, message)
            server.close()
            return True
        except Exception as e:
            print(f"Failed to send e-mail to {to_user}!\n\n" + str(e))
            return False

    def send_to_all(self, subject, body):
        to_user = 'ALL'
        # Sends experiment related messages to All of IPI
        body = f"Hi All,\n\nThe following is an automated message from {self.__sys_name}:\n\n{body}\n\nThanks!"
        message = "\r\n".join([f"From: {self.__from_addr}", f"To: {self.__all_addr}",f"Subject: {subject}", "", body])
        return self.__send_email(message, to_user)

    def send_to_admin(self, subject, body):
        to_user = 'ADMIN'
        #sends maintinence related messages to Admin
        body = f"Hi Admin!\n\n{self.__sys_name} is currently having" \
               f" the following issues that need to be addressed:\n\n{body}\n\nThanks!"
        message = "\r\n".join([f"From: {self.__from_addr}", f"To: {self.__admin_addr}", f"Subject: {subject}", "", body])
        return self.__send_email(message, to_user)



import os

class InputParser(object):

    def __init__(self, quit_command=['q', 'quit']):
        self._quit_cmd = quit_command
        self.prompt = []
        self.defaultinput = []
        self.previousvals = []
        self.inputbounds = []
        self.userinput = []
        self.cmdflags = []
        self.modifiers = []

    def __call__(self, inputarray):
        for val in inputarray:
            if val.lower() in self._quit_cmd:
                break
            isfloat = self.__is_float(val)
            ispath = self.__is_path(val)
            if ('-' not in val and not isfloat) or ispath:
                self.modifiers.append(val)
            elif '-' in val[0] and not isfloat:
                self.cmdflags.append(val.lower())
            else:
                try:
                    self.userinput.append(float(val))
                except ValueError:
                    print("Invalid Command Parameters!")
                    return False
        return True

    def clear(self):
        self.prompt = []
        self.defaultinput = []
        self.previousvals = []
        self.inputbounds = []
        self.userinput = []
        self.cmdflags = []
        self.modifiers = []
        return True

    def getparams(self):
        returnvals = [None] * len(self.prompt)
        if len(self.userinput) > len(returnvals):
            self.userinput[:len(returnvals)-1]
        for i in range(len(self.userinput)):
            if i >= len(returnvals) or i >= len(self.userinput):
                break
            returnvals[i] = self.userinput[i]
        for j in range(len(self.prompt)):
            inval = ''
            if returnvals[j] is not None:
                if float(returnvals[j]) < float(self.inputbounds[j][0]):
                    returnvals[j] = float(self.inputbounds[j][0])
                elif float(returnvals[j]) > float(self.inputbounds[j][1]):
                    returnvals[j] = float(self.inputbounds[j][1])
                if len(self.previousvals) == len(self.defaultinput) and len(self.defaultinput) > 0:
                    self.defaultinput = self.previousvals[:]
                    print(self.prompt[j] + " [" + str(self.defaultinput[j]) + "]: " + str(returnvals[j]))
            else:
                numeric = False
                while not numeric:
                    if len(self.previousvals) == len(self.defaultinput) and len(self.defaultinput) > 0:
                        self.defaultinput = self.previousvals[:]
                    inval = input(self.prompt[j] + " [" + str(self.defaultinput[j]) + "]: ")
                    if inval == '':
                        inval = self.defaultinput[j]
                    elif inval in self._quit_cmd:
                        print("Operation Aborted!")
                        return False
                    try:
                        temp = float(inval)
                        numeric = True
                    except ValueError:
                        numeric = False
                        continue
                    if float(inval) < float(self.inputbounds[j][0]):
                        returnvals[j] = self.inputbounds[j][0]
                    elif float(inval) > float(self.inputbounds[j][1]):
                        returnvals[j] = self.inputbounds[j][1]
                    else:
                        returnvals[j] = inval
        self.userinput = returnvals
        return returnvals

    def __is_float(self, input):
        try:
            num = float(input)
        except ValueError:
            return False
        return True

    def __is_int(self, input):
        try:
            num = int(input)
        except ValueError:
            return False
        return True

    def __is_path(self, input):
        path_check = [os.path.exists(input), os.access(os.path.dirname(input), os.W_OK)]
        return True if True in path_check else False



import numpy as np
import scipy


def lambertw(val):
    i = 0
    p = e = t = w = 0.00000
    eps = 4.0e-16 #eps = desiredprecision
    if val < -0.36787944117144232159552377016146086:
        return -np.inf #should return failure
    if val == 0:
        return 0.0
    #get initial approximation for iteration
    if val < 1.0: #series near 0
        p = np.sqrt(2.0 * (2.7182818284590452353602874713526625 * val+1.0))
        w = -1.0+p-p * p / 3.0+11.0 / 72.0 * p * p * p
    else: #asymptotic
        w = np.log(val)
    if val > 3.0:
        w -= np.log(w)
    while i < 20:  # Halley loop
        e = np.exp(w)
        t = w * e - val
        t /= e * (w + 1.0) - 0.5 * (w + 2.0) * t / (w + 1.0)
        w -= t
        if np.abs(t) < eps * (1.0 + np.abs(w)):
            return w  # rel - abs error
        i += 1
    # never gets here
    return np.inf


def calc_derivative(x_list, y_list):
    x_out = []
    y_out = []
    for i in range(len(x_list)):
        if i == 0:
            x_out.append(x_list[i])
            y_out.append((y_list[i+1] - y_list[i]) / (x_list[i+1] - x_list[i]))
        else:
            x_out.append(x_list[i])
            y_out.append((y_list[i] - y_list[i-1]) / (x_list[i] - x_list[i-1]))
    return x_out, y_out


def calc_integral(x_list, y_list):
	# require x for future implementations that may use dX
    x_out = []
    y_out = []
    for i in range(len(x_list)):
        if i == 0:
            x_out.append(x_list[i])
            y_out.append(y_list[i])
        else:
            x_out.append(x_list[i])
            y_out.append(y_list[i] + y_out[i-1])
    return x_out, y_out


def sav_golay_filter(data_vector, window_length=15, polyorder=3, deriv=0, delta=1.0, axis=-1, mode='interp', cval=0.0):
    return scipy.signal.savgol_filter(data_vector, window_length, polyorder, deriv, delta, axis, mode, cval)



import numpy as np
from scipy import stats
import ast
import os
import json


class init():
    def __init__(self):
        self.__data = Data()
        self.matrix = self.__data.matrix
        self.plot_limits = self.__data.plot_limits
        self.directories = Directories()

    def new_buffer(self):
        return Buffer()

    def new_matrix(self):
        return Data().matrix


class _SharedDCO(object):
    def __init__(self):
        self.generic_array_set_method = self.__generic_array_set_method
        self.generic_cat_list_set_method = self.__generic_cat_list_set_method

    @staticmethod
    def __generic_array_set_method(input_var):
        in_type = type(input_var)
        if isinstance(input_var, (np.ndarray, np.generic)) or in_type is list or in_type is tuple:
            return np.array(input_var).flatten()
        elif in_type is int or in_type is float:
            return np.array([input_var])
        elif in_type is str and len(input_var) > 3:
            try:
                conv_input = ast.literal_eval(input_var)
                conv_type = type(conv_input)
                if conv_type is list or conv_type is tuple:
                    return np.array(conv_input)
                elif conv_type is float or conv_type is int:
                    return np.array([input_var])
            except Exception as e:
                # not useable data pass to final error raise
                pass
        raise TypeError('Input variable is not int, float, np.array, list, tuple, or string representation thereof!')

    @staticmethod
    def __generic_cat_list_set_method(input_var):
        in_type = type(input_var)
        if isinstance(input_var, (np.ndarray, np.generic)) or in_type is list or in_type is tuple:
            return _SharedDCO.__flatten(input_var)
        elif in_type is int or in_type is float:
            return [str(input_var)]
        elif in_type is str and len(input_var) > 3:
            try:
                conv_input = ast.literal_eval(input_var)
                conv_type = type(conv_input)
                if conv_type is list or conv_type is tuple:
                    return _SharedDCO.__flatten(conv_input)
                else:
                    return [str(input_var)]
            except Exception as e:
                # not useable data pass to final error raise
                pass
        raise TypeError('Input variable is not int, float, np.array, list, tuple, or string representation thereof!')

    @staticmethod
    def __flatten(str_matrix):
        rt = []
        for i in str_matrix:
            if isinstance(i, list):
                rt.extend(_SharedDCO.__flatten(i))
            else:
                rt.append(str(i))
        return rt


class Data(object):
    def __init__(self):
        self.matrix = self.__Matrix()
        self.plot_limits = self._PlotLimits(self.matrix)

    class __Matrix(object):
        def __init__(self):
            self.__buffer_list = []

        def __len__(self):
            return len(self.__buffer_list)

        def get(self) -> list:
            return self.__buffer_list

        def set(self, data_matrix: list):
            if type(data_matrix) is not list or len(data_matrix) == 0:
                raise ValueError(f"Expecting 'List' type with length > 0, received: {type(data_matrix)}")
            self.__buffer_valid_check(data_matrix[0])
            self.__buffer_list = data_matrix

        def length(self) -> int:
            return len(self.__buffer_list)

        def buffer(self, buffer_number: int):
            return self.get_buffer_by_number(buffer_number)

        def get_buffer_by_number(self, buffer_number: int):
            return self.__buffer_list[self.__buffer_number_to_idx(buffer_number)]

        def set_buffer_by_number(self, buffer, buffer_number: int):
            self.__buffer_number_valid_check(buffer_number)
            self.__buffer_valid_check(buffer)
            self.__buffer_list[self.__buffer_number_to_idx(buffer_number)] = buffer

        def add_buffer(self, buffer):
            self.__buffer_valid_check(buffer)
            self.__buffer_list.append(buffer)

        def append_new_buffer(self):
            self.__buffer_list.append(Buffer())

        def remove_buffer_by_number(self, buffer_number: int):
            buffer_idx = buffer_number - 1
            self.__buffer_number_valid_check(buffer_number)
            del self.__buffer_list[buffer_idx]

        def clear(self):
            self.__buffer_list = []

        def shortest_x_length(self) -> int:
            shortest = np.inf
            for i in range(len(self.__buffer_list)):
                temp_len = self.__buffer_list[i].data.x.length()
                if temp_len < shortest:
                    shortest = temp_len
            return shortest

        def longest_x_length(self) -> int:
            longest = -np.inf
            for i in range(len(self.__buffer_list)):
                temp_len = len(self.__buffer_list[i].data.x)
                if temp_len > longest:
                    longest = temp_len
            return longest

        def __buffer_number_valid_check(self, buffer_number):
            buffer_idx = buffer_number - 1
            if buffer_idx not in range(len(self.__buffer_list)):
                buf_idx_lst = list(range(len(self.__buffer_list)))
                first = buf_idx_lst[0] + 1
                last = buf_idx_lst[-1] + 1
                raise ValueError(f"Buffer number: {buffer_number} is not in range of data matrix: {first} - {last}")

        @staticmethod
        def __buffer_number_to_idx(buffer_number: int) -> int:
            return buffer_number - 1

        @staticmethod
        def __buffer_valid_check(input_object):
            if not isinstance(input_object, Buffer):
                raise ValueError(f"Expecting PyVuka buffer objects, recieved: {type(input_object)}")

    class _PlotLimits(object):
        def __init__(self, matrix_instance):
            self.buffer_range = self._BufferRange()
            self.x_range = self._XYZRange()
            self.y_range = self._XYZRange()
            self.z_range = self._XYZRange()
            self.__matrix_save = None
            self.is_active = False
            self.matrix = matrix_instance

        def on(self):
            if self.is_active:
                return
            self.__matrix_save = self.__copy_matrix_data()  # faster than using copy.deepcopy

            if len(self.buffer_range.get()) > 0:
                firstbuffer, lastbuffer = [min(self.buffer_range.get()), max(self.buffer_range.get())]
            else:
                firstbuffer, lastbuffer = [1, self.matrix.length()]
            for i in range(firstbuffer, lastbuffer + 1):
                self.matrix.set_buffer_by_number(self.apply_to_buffer(self.matrix.buffer(i), self), i)
            self.is_active = True

        def __copy_matrix_data(self):
            save_list = []
            for i in range(1, len(self.matrix) + 1):
                save_dict = {}
                save_dict['x'] = list(self.matrix.buffer(i).data.x.get().astype(float))
                save_dict['xe'] = list(self.matrix.buffer(i).data.xe.get().astype(float))
                save_dict['y'] = list(self.matrix.buffer(i).data.y.get().astype(float))
                save_dict['ye'] = list(self.matrix.buffer(i).data.ye.get().astype(float))
                save_dict['z'] = list(self.matrix.buffer(i).data.z.get().astype(float))
                save_dict['ze'] = list(self.matrix.buffer(i).data.ze.get().astype(float))
                save_list.append(save_dict)
            return json.loads(json.dumps(save_list))

        def off(self):
            if not self.is_active:
                return
            for i in range(1, len(self.__matrix_save) + 1):
                if i in range(min(self.buffer_range.get()), max(self.buffer_range.get()) + 1):
                    self.matrix.buffer(i).data.x.set(self.__matrix_save[i - 1]['x'])
                    self.matrix.buffer(i).data.xe.set(self.__matrix_save[i - 1]['xe'])
                    self.matrix.buffer(i).data.y.set(self.__matrix_save[i - 1]['y'])
                    self.matrix.buffer(i).data.ye.set(self.__matrix_save[i - 1]['ye'])
                    self.matrix.buffer(i).data.z.set(self.__matrix_save[i - 1]['z'])
                    self.matrix.buffer(i).data.ze.set(self.__matrix_save[i - 1]['ze'])
            self.is_active = False
            self.__matrix_save = None

        @staticmethod
        def apply_to_buffer(buffer_object, plot_limits):
            new_buffer = buffer_object  # removed copy.deepcopy
            x = new_buffer.data.x.get()
            xe = new_buffer.data.xe.get()
            y = new_buffer.data.y.get()
            ye = new_buffer.data.ye.get()
            z = new_buffer.data.z.get()
            ze = new_buffer.data.ze.get()

            x_range = plot_limits.x_range.get()
            y_range = plot_limits.y_range.get()
            z_range = plot_limits.z_range.get()
            idx_out = [i for i, j in enumerate(x) if not min(x_range) - 1 <= i <= max(x_range) - 1] if len(x_range) > 0 else [] + \
                [i for i, j in enumerate(y) if not min(y_range) - 1 <= j <= max(y_range) - 1] if len(y_range) > 0 else [] + \
                [i for i, j in enumerate(z) if not min(z_range) - 1 <= j <= max(z_range) - 1] if len(z_range) > 0 else []

            if len(idx_out) == 0:
                return buffer_object

            idx_out = list(set(sorted(idx_out)))
            all_data = [x, y, z, xe, ye, ze]
            for i, data_vector in enumerate(all_data):
                if len(data_vector) > 0:
                    if max(idx_out) <= len(data_vector):
                        all_data[i] = np.delete(data_vector, idx_out)

            x, y, z, xe, ye, ze = all_data

            new_buffer.data.x.set(x)
            new_buffer.data.xe.set(xe)
            new_buffer.data.y.set(y)
            new_buffer.data.ye.set(ye)
            new_buffer.data.z.set(z)
            new_buffer.data.ze.set(ze)
            return new_buffer

        class _base_range(object):
            def __init__(self):
                self.__base = tuple([])

            def get(self) -> tuple:
                return self.__base

            def set(self, user_input: iter):
                if type(user_input) is str:
                    user_input = ast.literal_eval(user_input)
                if not is_iterable(user_input):
                    raise ValueError(f"Parameter: {user_input} is not iterable. Expecting tuple containing 2 numbers")
                self.__base = tuple(user_input[:2])

            def max(self):
                return max(self.__base) if len(self.__base) > 0 else None

            def min(self):
                return min(self.__base) if len(self.__base) > 0 else None

        class _BufferRange(_base_range):
            def __init__(self):
                super(Data._PlotLimits._BufferRange, self).__init__()
                # self.__base = tuple([int(x) for x in self.__base].sort()[:2]) if len(self.__base) >= 2 else None

        class _XYZRange(_base_range):
            def __init__(self):
                super(Data._PlotLimits._XYZRange, self).__init__()
                # self.__base = tuple([float(x) for x in self.__base].sort()[:2]) if len(self.__base) >= 2 else None


class Directories(object):
    def __init__(self):
        self.working = self.__Working()
        self.output = self.__Output()

    def __str__(self):
        return f"Working Directory: {self.working}\nOutput Directory: {self.output}"

    class _base_class(object):
        def __init__(self):
            self.__base = ''

        def get(self) -> str:
            return self.__base

        def set(self, dir_name):
            try:
                if not os.path.exists(dir_name):
                    os.makedirs(dir_name)
            except Exception as e:
                raise ValueError(f'Supplied directory path: {dir_name} does not exist and cannot be created!')
            self.__base = str(dir_name)

    class __Working(_base_class):
        pass

    class __Output(_base_class):
        pass


class Buffer(object):
    def __init__(self):
        self.data = self.__BaseData()
        self.category = self.__BaseCategory()
        self.model = self.__BaseData()
        self.residuals = self.__BaseData()
        self.instrument_response = self.__BaseData()
        self.fit = self.__Fit()
        self.plot = self.__Plot()
        self.comments = self.__Comments()
        self.meta_dict = {}

    class __BaseData(object):
        def __init__(self):
            self.x = self._base_array()
            self.xe = self._base_array()
            self.y = self._base_array()
            self.ye = self._base_array()
            self.z = self._base_array()
            self.ze = self._base_array()
            self.color = self._base_color()
            self.is_visible = True
            self.weight = self._base_weight()

        def show(self, yes=True) -> bool:
            self.is_visible = True if yes is True else False

        def hide(self, yes=True) -> bool:
            self.is_visible = False if yes is True else True

        class _base_weight(object):
            def __init__(self):
                self.__weight = 2

            def get(self) -> int:
                return self.__weight

            def set(self, user_input: int):
                self.__weight = user_input

        class _base_array(object):
            def __init__(self):
                self.__base = np.array([])

            def __len__(self):
                self.length()

            def get(self) -> np.array:
                return self.__base

            def get_sorted_ascending(self) -> np.array:
                return np.sort(self.__base)

            def get_sorted_decending(self) -> np.array:
                return np.sort(self.__base)[::-1]

            def set(self, user_input: iter):
                self.__base = _SharedDCO().generic_array_set_method(user_input)

            def append(self, value):
                self.__base = np.append(self.__base, value)

            def clear(self):
                self.__base = np.array([])

            def set_sorted_ascending(self, user_input: iter = None):
                if user_input is None or len(self.__base) == 0:
                    raise ValueError("No data to sort!")
                elif user_input is not None:
                    self.set(user_input)
                self.__base = self.get_sorted_ascending()

            def set_sorted_decending(self, user_input: iter = None):
                if user_input is None or len(self.__base) == 0:
                    raise ValueError("No data to sort!")
                elif user_input is not None:
                    self.set(user_input)
                self.__base = self.get_sorted_decending()

            def set_random(self, minimum: float = 0, maximum: float = 100, num_pts: int = -1):
                if num_pts > 0:
                    self.__base = np.random.uniform(low=minimum, high=maximum, size=(num_pts))
                elif len(self.__base) > 0:
                    self.__base = np.random.uniform(low=minimum, high=maximum, size=(len(self.__base)))
                else:
                    raise ValueError("No number of random values is defined!")

            def set_zeros(self, num_pts: int = -1) -> np.array:
                if num_pts < 0:
                    self.__base = np.zeros(num_pts)
                elif len(self.__base) > 0:
                    self.__base = np.zeros(len(self.__base))
                else:
                    raise ValueError("No number of zeros defined!")

            def average(self) -> float:
                return float(np.nanmean(self.__base)) if len(self.__base) > 0 else None

            def stdev(self) -> float:
                return float(np.nanstd(self.__base)) if len(self.__base) > 0 else None

            def range(self) -> tuple:
                data = [x for x in self.__base if np.isfinite(x) and not np.isnan(x)]
                return tuple([np.nanmin(data), np.nanmax(data)]) if len(data) > 1 else None

            def median(self) -> float:
                return float(np.nanmedian(self.__base)) if len(self.__base) > 0 else None

            def mode(self) -> tuple:
                '''Returns tuple of array_of_modal_values, array_of_mode_counts'''
                return stats.mode(self.__base) if len(self.__base) > 0 else None

            def min(self) -> float:
                return float(np.nanmin(self.__base)) if len(self.__base) > 0 else None

            def max(self):
                return np.nanmax(self.__base) if len(self.__base) > 0 else None

            def length(self) -> int:
                return len(self.__base)

            def sum(self) -> float:
                return float(np.nansum(self.__base)) if len(self.__base) > 0 else None

            def cumsum(self) -> np.array:
                return np.nancumsum(self.__base) if len(self.__base) > 0 else None

            def product(self) -> np.array:
                return np.nanprod(self.__base) if len(self.__base) > 0 else None

            def cumproduct(self) -> np.array:
                return np.nancumprod(self.__base) if len(self.__base) > 0 else None

            def clean_nan_inf(self):
                self.__base = np.nan_to_num(self.__base)

            def nearest_index_to_value(self, value: float) -> int:
                subarray = np.abs(np.array(self.__base) - float(value))
                return int(subarray.argmin())

            def value_at_index(self, index: int) -> float:
                return float(self.__base[index]) if index <= len(self.__base) else None

        class _base_color(object):
            def __init__(self):
                self.__color_val = '#000000'  # black

            def set(self, color):
                self.__color_val = color

            def get(self):
                if len(str(self.__color_val)) > 0 and str(self.__color_val)[0] == '(' and str(self.__color_val)[
                    -1] == ')':
                    return tuple(self.__color_val)
                return self.__color_val

    class __BaseCategory(object):
        def __init__(self):
            self.x = self._base_cat_array()
            self.y = self._base_cat_array()
            self.z = self._base_cat_array()

        class _base_cat_array(object):
            def __init__(self):
                self.__base = []

            def get(self) -> list:
                return self.__base

            def set(self, user_input: iter):
                self.__base = _SharedDCO().generic_cat_list_set_method(user_input)

            def append(self, value):
                if type(value) in [str, float, int]:
                    self.__base.append(str(value))
                elif is_iterable(value):
                    self.__base.extend([str(v) for v in value])
                else:
                    raise ValueError('Unexpected parameter type! Expecting: str, float, int, list, or tuple')

            def clear(self):
                self.__base = []

            def get_sorted_ascending(self) -> list:
                return list(sorted(self.__base))

            def get_sorted_decending(self) -> list:
                return list(sorted(self.__base, reverse=True))

            def set_sorted_ascending(self, user_input: iter = None):
                if user_input is None or len(self.__base) == 0:
                    raise ValueError("No data to sort!")
                elif user_input is not None:
                    self.set(user_input)
                self.__base = self.get_sorted_ascending()

            def set_sorted_decending(self, user_input: iter = None):
                if user_input is None or len(self.__base) == 0:
                    raise ValueError("No data to sort!")
                elif user_input is not None:
                    self.set(user_input)
                self.__base = self.get_sorted_decending()

            def range(self) -> tuple:
                data = [x for x in self.__base if np.isfinite(x) and not np.isnan(x)]
                return tuple([data[0], data[-1]]) if len(data) > 1 else None

            def mode(self) -> tuple:
                '''Returns tuple of array_of_modal_values, array_of_mode_counts'''
                return stats.mode(self.__base) if len(self.__base) > 0 else None

            def length(self) -> int:
                return len(self.__base)

            def indcies_of_value(self, value: str) -> list:
                return [i for i, entry in enumerate(self.__base) if entry.strip().lower() == value.strip().lower()]

            def value_at_index(self, index: int) -> str:
                return self.__base[index] if index <= len(self.__base) else None

    class __Fit(object):
        def __init__(self):
            self.function = self._base_str()
            self.function_index = self._base_list()
            self.parameter = self._base_list()
            self.parameter_error = self._base_list()
            self.chisq = self._base_float()
            self.rsq = self._base_float()
            self.link = self._base_list()
            self.free = self._base_list()
            self.use_error_weighting = True
            self.fit_failed = False
            self.fit_failed_reason = self._base_str()

        class _base_str(object):
            def __init__(self):
                self.__str_val = ''

            def set(self, value: str):
                self.__str_val = str(value)

            def get(self) -> str:
                return self.__str_val

        class _base_list(object):
            def __init__(self):
                self.__base = []

            def __len__(self):
                self.length()

            def length(self):
                return len(self.__base)

            def get(self):
                return self.__base

            def set(self, input_list: iter):
                if is_iterable(input_list):
                    self.__base = list(input_list)
                else:
                    raise ValueError(f"Input is {type(input_list)}, not List or Tuple!")

        class _base_float(object):
            def __init__(self):
                self.__base = 0.00

            def get(self):
                return self.__base

            def set(self, value: float):
                try:
                    self.__base = float(value)
                except Exception as e:
                    raise ValueError(f"Input is {type(value)}, not the expected Float type!")

    class __Plot(object):
        def __init__(self):
            self.type = self._BaseStr()
            self.title = self._BaseStr()
            self.series = self.__Series()
            self.axis = self.__Axis()
            self.polygons = self.__Polygons()
            self.use_weighted_residuals = False

        class _BaseStr(object):
            def __init__(self):
                self.__str_val = ''

            def set(self, value: str):
                self.__str_val = str(value)

            def get(self) -> str:
                return self.__str_val

        class __Series(object):
            def __init__(self):
                self.name = self._base_str()
                self.color = self._base_color()
                self.type = self._base_str()
                self.weight = self._base_float()

            class _base_float(object):
                def __init__(self):
                    self.__float_val = 5

                def set(self, value: float):
                    self.__float_val = float(value)

                def get(self) -> float:
                    return self.__float_val

            class _base_str(object):
                def __init__(self):
                    self.__str_val = ''

                def set(self, value: str):
                    self.__str_val = str(value)

                def get(self) -> str:
                    return self.__str_val

            class _base_color(object):
                def __init__(self):
                    self.__color_val = '#FF0000'  # red

                def set(self, color):
                    self.__color_val = color

                def get(self):
                    if len(str(self.__color_val)) > 0 and \
                            str(self.__color_val)[0] == '(' \
                            and str(self.__color_val)[-1] == ')':
                        return tuple(self.__color_val)
                    return self.__color_val

            class _base_type(object):
                def __init__(self):
                    self.__type_val = '.'
                    self.__valid_types = self.return_valid_types()

                def set(self, point_type: str):
                    is_valid = False
                    point_type = point_type.lower()
                    if 'string' in point_type and '(' in point_type and point_type[-1] == ')':
                        point_type = f'${point_type.split("(")[-1][:-1]}$'
                        is_valid = True
                    else:
                        for k, v in self.__valid_types.items():
                            if point_type == k or point_type == v:
                                point_type = v
                                is_valid = True
                                break
                    self.__type_val = str(point_type) if is_valid else '.'

                def get(self) -> str:
                    return self.__type_val

                @staticmethod
                def return_valid_types():
                    valid_types = {'point': '.',
                                   'pixel': ',',
                                   'circle': 'o',
                                   'triangle_down': 'v',
                                   'triangle_up': '^',
                                   'triangle_left': '<',
                                   'triangle_right': '>',
                                   'tri_down': '1',
                                   'tri_up': '2',
                                   'tri_left': '3',
                                   'tri_right': '4',
                                   'octagon': '8',
                                   'square': 's',
                                   'pentagon': 'p',
                                   'plus (filled)': 'P',
                                   'star': '*',
                                   'hexagon1': 'h',
                                   'hexagon2': 'H',
                                   'plus': '+',
                                   'x': 'x',
                                   'x (filled)': 'X',
                                   'diamond': 'D',
                                   'thin_diamond': 'd',
                                   'vline': '|',
                                   'hline': '_',
                                   'tickleft': 0,
                                   'tickright': 1,
                                   'tickup': 2,
                                   'tickdown': 3,
                                   'caretleft': 4,
                                   'caretright': 5,
                                   'caretup': 6,
                                   'caretdown': 7,
                                   'string ex: @': 'string(@)'}
                    return valid_types

        class __Axis(object):
            def __init__(self):
                self.x = self._base_axis()
                self.y = self._base_axis()
                self.z = self._base_axis()

            class _base_axis(object):
                def __init__(self):
                    self.title = self.__base_title()
                    self.axis_scale = self.__base_scale()
                    self.range = self.__base_range()
                    self.lines = self.__base_lines()
                    self.peaks = self.__base_nparray()
                    self.peak_bounds = self.__base_list()
                    self.integrals = self.__base_list()
                    self.label = self.__base_label()

                class __base_label(object):
                    def __init__(self):
                        self.size = self.__size()
                        self.is_visible = self.show()

                    def show(self, yes=True) -> bool:
                        self.is_visible = yes
                        return self.is_visible

                    def hide(self, yes=True) -> bool:
                        self.is_visible = False if yes is True else True
                        return self.is_visible

                    class __size(object):
                        def __init__(self):
                            self.__size = 10

                        def get(self) -> np.array:
                            return self.__size

                        def set(self, user_input):
                            if type(user_input) in [int, float]:
                                self.__size = int(user_input) if int(user_input) > 0 else 8
                            else:
                                raise ValueError(f"Invalid parameter type: {type(user_input)}; Expecting int type!")

                class __base_nparray(object):
                    def __init__(self):
                        self.__base_array = np.array([])
                        self.is_visible = self.hide()

                    def append(self, constant):
                        if not is_iterable(constant):
                            constant = [float(constant)]
                        self.__base_array = np.append(self.__base_array, constant, axis=0)

                    def get(self) -> np.array:
                        return self.__base_array

                    def set(self, value_list: list):
                        if type(value_list) is str:
                            lines = ast.literal_eval(value_list)
                        if type(value_list) in [int, float]:
                            lines = [value_list]
                        if is_iterable(value_list):
                            self.__base_array = np.array(value_list)
                        else:
                            raise ValueError(f"Invalid parameter type: {type(value_list)}; Expecting List type!")

                    def show(self, yes=True) -> bool:
                        self.is_visible = yes
                        return self.is_visible

                    def hide(self, yes=True) -> bool:
                        self.is_visible = False if yes is True else True
                        return self.is_visible

                class __base_lines(object):
                    def __init__(self):
                        self.__base_array = np.array([])
                        self.is_visible = self.show()
                        self.color = self._base_color()
                        self.weight = self._base_weight()
                        self.outline = self._outline()
                        self.line_style = self._line_style()

                    def append(self, constant):
                        if not is_iterable(constant):
                            constant = [float(constant)]
                        self.__base_array = np.append(self.__base_array, constant, axis=0)

                    def get(self) -> np.array:
                        return self.__base_array

                    def set(self, value_list: list):
                        if type(value_list) is str:
                            lines = ast.literal_eval(value_list)
                        if type(value_list) in [int, float]:
                            lines = [value_list]
                        if is_iterable(value_list):
                            self.__base_array = np.array(value_list)
                        else:
                            raise ValueError(f"Invalid parameter type: {type(value_list)}; Expecting List type!")

                    def show(self, yes=True) -> bool:
                        self.is_visible = yes
                        return self.is_visible

                    def hide(self, yes=True) -> bool:
                        self.is_visible = False if yes is True else True
                        return self.is_visible

                    class _outline(object):
                        def __init__(self):
                            self.is_visible = True

                        def hide(self):
                            self.is_visible = False

                        def show(self):
                            self.is_visible = True

                    class _base_weight(object):
                        def __init__(self):
                            self.__weight = 2

                        def get(self) -> int:
                            return self.__weight

                        def set(self, user_input: int):
                            self.__weight = user_input

                    class _line_style(object):
                        def __init__(self):
                            self.__style = '--'
                            self.__valid_types = ['-', '--', '-.', ':', 'None']

                        def get(self) -> int:
                            return self.__style

                        def set(self, user_input: int):
                            if user_input in self.__valid_types:
                                self.__style = user_input
                            else:
                                print(f'No change to x-line or y-line line style. Valid types: {self.__valid_types}')

                    class _base_color(object):
                        def __init__(self):
                            self.__color_val = '#808080'  # gray

                        def set(self, color):
                            self.__color_val = color

                        def get(self):
                            if len(str(self.__color_val)) > 0 and str(self.__color_val)[0] == '(' and \
                                    str(self.__color_val)[-1] == ')':
                                return tuple(self.__color_val)
                            return self.__color_val

                class __base_list(object):
                    def __init__(self):
                        self.__base = []
                        self.is_visible = self.hide()

                    def add(self, bounds: tuple):
                        if is_iterable(bounds):
                            if is_iterable(bounds[0]):
                                for b in bounds:
                                    self.__base.append(tuple(list(b)[:2]))
                            else:
                                self.__base.append(tuple(list(bounds)[:2]))
                        else:
                            raise ValueError(f"Invalid parameter type:{type(bounds)}; " +
                                             f"Expecting tuple or list of tuples")

                    def get(self) -> list:
                        return self.__base

                    def set(self, bounds: list):
                        to_set = []
                        if is_iterable(bounds):
                            if is_iterable(bounds[0]):
                                for b in bounds:
                                    to_set.append(tuple(list(b)[:2]))
                            else:
                                to_set = [tuple(list(bounds)[:2])]
                            self.__base = to_set
                        else:
                            raise ValueError(f"Unexpected data type: {type(bounds)}. " +
                                             f"Expecting list of tuples: [(lower_bound, upper_bound),...]!")

                    def show(self, yes=True) -> bool:
                        self.is_visible = yes
                        return self.is_visible

                    def hide(self, yes=True) -> bool:
                        self.is_visible = False if yes is True else True
                        return self.is_visible

                class __base_range(object):
                    def __init__(self):
                        self.__range = tuple([])

                    def set(self, range: tuple):
                        if is_iterable(range):
                            self.__range = tuple(list(range)[:2])
                        else:
                            raise ValueError(f"Type: {type(range)} is not valid. Expecting: tuple or list")

                    def get(self) -> str:
                        data = [x for x in self.__range if np.isfinite(x) and not np.isnan(x)]
                        return tuple([np.nanmin(data), np.nanmax(data)]) if len(data) > 1 else tuple([])

                class __base_scale(object):
                    def __init__(self):
                        self.__scale_base = 'linear'
                        self.__valid_types = self.return_valid_types()

                    def set(self, scale_type: str):
                        if scale_type.lower() in self.__valid_types:
                            self.__scale_base = scale_type.lower()
                        else:
                            raise ValueError(f"Type: {scale_type} is not valid. Expecting: {str(self.__valid_types)}")

                    def get(self) -> str:
                        return self.__scale_base

                    @staticmethod
                    def return_valid_types():
                        return ["linear", "log", "symlog", "logit"]

                class __base_title(object):
                    def __init__(self):
                        self.__base = ''

                    def set(self, title: str):
                        self.__base = str(title)

                    def get(self) -> str:
                        return self.__base

        class __Polygons(object):
            def __init__(self):
                self.__base_val = []
                self.is_visible = self.hide()

            def get(self, index=-1) -> np.array:
                if index == -1:
                    return self.__base_val
                else:
                    return self.__base_val[index]

            def add_verticies(self, vertex, polygon_index=0):
                self.__base_val[polygon_index] = np.append(self.__base_val[polygon_index], vertex, axis=0)

            def set(self, verticies_list, index=-1):
                if index == -1:
                    if type(verticies_list[0]) is list:
                        self.__base_val = verticies_list
                    else:
                        self.__base_val = [verticies_list]
                else:
                    self.__base_val[index] = verticies_list

            def add_polygon(self, verticie_list: list):
                self.__base_val.append(np.array(verticie_list))

            def remove_polygon(self, index: int):
                del self.__base_val[index]

            def clear(self):
                self.__base_val = []

            def show(self, yes=True) -> bool:
                self.is_visible = yes
                return self.is_visible

            def hide(self, yes=True) -> bool:
                self.is_visible = False if yes is True else True
                return self.is_visible

    class __Comments(object):
        def __init__(self):
            self.comments = []

        def length(self) -> int:
            return len(self.comments)

        def get(self) -> list:
            return self.comments

        def set(self, user_comment: iter):
            if not is_iterable(user_comment):
                user_comment = [str(user_comment)]
            self.comments = [str(x) for x in user_comment]

        def add(self, user_comment: str):
            if is_iterable(user_comment):
                user_comment = str(user_comment)[1:-1]
            self.comments.append(user_comment)

        def remove_comment_by_index(self, index: int):
            if index not in range(self.length()):
                raise ValueError(f"Index: {index} is out of range: 0 - {self.length()}")
            del self.comments[index]

        def all_as_string(self) -> str:
            str_out = ''
            is_append = False
            for c in self.comments:
                is_append = True
                str_out += c + ' | '
            if is_append:
                str_out = str_out[:-3]
            return str_out

    def to_dict(self):
        output = {'data_x': self.data.x.get(),
                  'data_xe': self.data.xe.get(),
                  'data_y': self.data.y.get(),
                  'data_ye': self.data.ye.get(),
                  'data_z': self.data.z.get(),
                  'data_ze': self.data.ze.get(),
                  'category_x': self.category.x.get(),
                  'category_y': self.category.y.get(),
                  'category_z': self.category.z.get(),
                  'model_x': self.model.x.get(),
                  'model_y': self.model.y.get(),
                  'model_z': self.model.z.get(),
                  'residuals_x': self.residuals.x.get(),
                  'residuals_y': self.residuals.y.get(),
                  'residuals_z': self.residuals.z.get(),
                  'instrument_response_x': self.instrument_response.x.get(),
                  'instrument_response_y': self.instrument_response.y.get(),
                  'instrument_response_z': self.instrument_response.z.get(),
                  'plot_title': self.plot.title.get(),
                  'plot_type': self.plot.type.get(),
                  'plot_polygons': self.plot.polygons.get(),
                  'plot_use_weighted_residuals': self.plot.use_weighted_residuals,
                  'plot_series_name': self.plot.series.name.get(),
                  'plot_series_color': self.plot.series.color.get(),
                  'plot_series_type': self.plot.series.type.get(),
                  'plot_series_weight': self.plot.series.weight.get(),
                  'plot_x_title': self.plot.axis.x.title.get(),
                  'plot_x_type': self.plot.axis.x.axis_scale.get(),
                  'plot_x_integrals': self.plot.axis.x.integrals.get(),
                  'plot_x_lines': self.plot.axis.x.lines.get(),
                  'plot_x_peak_bounds': self.plot.axis.x.peak_bounds.get(),
                  'plot_x_peaks': self.plot.axis.x.peaks.get(),
                  'plot_x_range': self.plot.axis.x.range.get(),
                  'plot_y_title': self.plot.axis.y.title.get(),
                  'plot_y_type': self.plot.axis.y.axis_scale.get(),
                  'plot_y_integrals': self.plot.axis.y.integrals.get(),
                  'plot_y_lines': self.plot.axis.y.lines.get(),
                  'plot_y_peak_bounds': self.plot.axis.y.peak_bounds.get(),
                  'plot_y_peaks': self.plot.axis.y.peaks.get(),
                  'plot_y_range': self.plot.axis.y.range.get(),
                  'plot_z_title': self.plot.axis.z.title.get(),
                  'plot_z_type': self.plot.axis.z.axis_scale.get(),
                  'plot_z_integrals': self.plot.axis.z.integrals.get(),
                  'plot_z_lines': self.plot.axis.z.lines.get(),
                  'plot_z_peak_bounds': self.plot.axis.z.peak_bounds.get(),
                  'plot_z_peaks': self.plot.axis.z.peaks.get(),
                  'plot_z_range': self.plot.axis.z.range.get(),
                  'comments': self.comments.all_as_string(),
                  }
        output = {k: str(v) for k, v in output.items() if str(v) not in ['[]', '()', 'None', '']}
        return output


def is_iterable(element):
    if type(element) is str:
        return False
    try:
        iterator = iter(element)
    except TypeError:
        return False
    else:
        return True



import setuptools
import PyVuka.pyvuka as pvk

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name=pvk.__app_name__,
    version=pvk.__version__,
    author=pvk.__author__,
    author_email=pvk.__email__,
    description=pvk.__description__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bostonautolytics/pyvuka",
    packages=setuptools.find_packages(),
    install_requires=["asteval>=0.9.18", "chardet>=3.0.4", "lmfit>=1.0.0", "matplotlib>=3.1.3", "numpy>=1.18.1",
                      "pack64>=2.0.1", "Pillow>=7.0.0", "psutil>=5.6.7", "PyQt5>=5.14.1", "scipy>=1.3.1",
                      "xlrd>=1.2.0", "XlsxWriter>=1.2.7"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Free For Educational Use",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


#!/usr/bin/env python
#title           :app.py
#description     :This file is used to start the PyVuka commandline program
#                 through a python import with clear syntax
#usage           :import PyVuka.app as app
#                 app.start()
#python_version  :3.7
#==============================================================================
from . import pyvuka


def start():
    pyvuka.start()


